from downstream_clustering.kmeans_based import features_kmeans, features_n_kmeans
from munkres import Munkres, print_matrix
from sklearn import metrics
import numpy as np
import os
# from EMAN2 import EMNumPy,Averagers
import stat
import shutil
from tqdm import tqdm
import multiprocessing
from functools import partial
from PIL import Image
import torch
from torchvision.utils import save_image
import pandas as pd
from cryoemdata.data_preprocess.mrc_preprocess import mrcs_resize
from models.get_transformers import filter
import cv2
import pickle
from cryoemdata.data_preprocess.mrc_preprocess import mrcs_to_int8

# from cryoemdata.e2classaverage import class_average_withali, class_average
# from EMAN2 import Transform

class Clustering_evaluator(object):
    def __init__(self, particles_path, result_path_dir, k, particles_len,particles_id=None, labels_p=None, labels_t=None,
                 features=None):
        self.result_path = result_path_dir
        self.particles_path = particles_path
        if particles_id is None:
            self.particles_id = range(particles_len)
        else:
            self.particles_id = particles_id
        self.labels_p = labels_p
        self.labels_t = labels_t
        self.k = k
        self.features = features

    def clustering(self, features,k=None):
        if k==None:
            k=self.k
        self.features = features
        # labels_predict, centers, num_class = features_kmeans(features,k=self.k)
        labels_predict, centers, num_class = features_n_kmeans(features, k=k)
        self.labels_p = labels_predict
        self.centers = centers
        return labels_predict, centers, num_class

    def acc_nmi(self, labels_t):
        self.labels_t = labels_t
        label_same = best_map(self.labels_t, self.labels_p)
        count = np.sum(self.labels_t[:] == label_same[:])
        acc = count.astype(float) / (len(self.labels_t))
        nmi = metrics.normalized_mutual_info_score(self.labels_t, label_same)
        return acc, nmi

    def save_labels(self, epoch, particles_id=None,save_path_dir_name=None):
        if particles_id is None:
            ids = self.particles_id
        else:
            ids = particles_id
        col_name = ['clustering_results']
        df = pd.DataFrame(columns=col_name, data=self.labels_p, index=ids)
        if save_path_dir_name is None:
            labels_predict_path = self.result_path + '/clustering_predict/epoch_' + str(epoch)
        else:
            labels_predict_path = os.path.join(self.result_path, save_path_dir_name, '/epoch_' + str(epoch))
        # labels_predict_path = self.result_path + '/clustering_predict/epoch_' + str(epoch)
        if not os.path.exists(labels_predict_path):
            os.makedirs(labels_predict_path)
        df.to_csv(labels_predict_path + '/clustering_predict.csv', encoding='utf-8')

        with open(labels_predict_path + '/output_tif_label.data', 'wb') as filehandle:
            pickle.dump(self.labels_p, filehandle)
        # with open(labels_predict_path + '/clustering_predict.data', 'wb') as filehandle:
        #     pickle.dump(self.labels_p, filehandle)

    # def save_particles_list(self):
    #     col_name = ['particles']
    #     df = pd.DataFrame(columns=col_name, data=self.particles_path)
    #     labels_predict_path = self.result_path + '/clustering_predict/'
    #     if not os.path.exists(labels_predict_path):
    #         os.makedirs(labels_predict_path)
    #     df.to_csv(labels_predict_path + '/particles_list.csv', encoding='utf-8', index=False)

        # with open(labels_predict_path + '/clustering_predict.data', 'wb') as filehandle:
        #     pickle.dump(self.labels_p, filehandle)

    def save_averages(self, epoch):
        my_save_averages(epoch, self.result_path, self.averages, self.original_averages)
        # averages_predict_path = self.result_path + '/clustering_predict/epoch_' + str(epoch)
        # if not os.path.exists(averages_predict_path):
        #     os.makedirs(averages_predict_path)
        # with open(averages_predict_path + '/averages.data', 'wb') as filehandle:
        #     pickle.dump(self.original_averages, filehandle)
        # if np.ndim(self.averages) == 3:
        #     self.averages=np.expand_dims(self.averages, axis=1)
        # save_image(torch.tensor(self.averages), averages_predict_path + '/averages.png')
        # for i in range(self.averages.shape[0]):
        #     Image.fromarray(self.averages[i]).save(averages_predict_path+'/'+str(i)+'.pdf')

    def update_labels(self, labels_p=None, labels_t=None):
        self.labels_p = labels_p
        self.labels_t = labels_t

    def embedding_projection(self, tb_writer, labels_t=None, title=None, epoch=None):
        if self.labels_p.shape[0] > 20000:
            if labels_t is not None:
                sub_labels = labels_t[:20000]
            else:
                sub_labels = self.labels_p[:20000]
            sub_features = self.features[:20000]
        else:
            if labels_t is not None:
                sub_labels = labels_t
            else:
                sub_labels = self.labels_p
            sub_features = self.features
        tb_writer.add_embedding(mat=sub_features, metadata=sub_labels, global_step=epoch, tag=title)

    @torch.no_grad()
    def knn_classifier(self, train_features=None, train_labels=None, test_features=None, test_labels=None,
                       num_classes=None, k=5, T=1,mask_start_index=0):
        if num_classes is None:
            num_classes = self.k
        if train_labels is None:
            train_labels = self.labels_t
        if test_labels is None:
            test_labels = self.labels_t
        if train_features is None:
            train_features = self.features
        if test_features is None:
            test_features = self.features

        train_features = torch.from_numpy(train_features).cuda()
        train_labels = torch.from_numpy(np.asarray(train_labels)).cuda()
        test_features = torch.from_numpy(test_features).cuda()
        test_labels = torch.from_numpy(np.asarray(test_labels)).cuda()

        top1, top5, total = 0.0, 0.0, 0
        train_features = train_features.t()
        num_test_images, num_chunks = test_labels.shape[0], int(test_labels.shape[0]/1000)
        imgs_per_chunk = num_test_images // num_chunks
        retrieval_one_hot = torch.zeros(k, num_classes).to(train_features.device)
        for idx in range(0, num_test_images, imgs_per_chunk):
            # get the features for test images
            features = test_features[
                       idx: min((idx + imgs_per_chunk), num_test_images), :
                       ]
            targets = test_labels[idx: min((idx + imgs_per_chunk), num_test_images)]
            batch_size = targets.shape[0]

            # calculate the dot product and compute top-k neighbors
            similarity = torch.mm(features, train_features)
            # similarity_np=similarity.cpu().numpy()
            # mm=torch.arange(idx, min((idx + imgs_per_chunk), num_test_images)).unsqueeze(-1)
            mask = torch.ones_like(similarity).scatter_(1, torch.arange(mask_start_index+idx, mask_start_index+min((idx + imgs_per_chunk),
                                                                                 num_test_images)).unsqueeze(-1).cuda(),
                                                        0)
            # aaaa = mask.cpu().numpy()
            similarity = similarity.masked_fill(mask == 0, 0)
            distances, indices = similarity.topk(k, largest=True, sorted=True)
            candidates = train_labels.view(1, -1).expand(batch_size, -1)
            retrieved_neighbors = torch.gather(candidates, 1, indices)

            retrieval_one_hot.resize_(batch_size * k, num_classes).zero_()
            retrieval_one_hot.scatter_(1, retrieved_neighbors.view(-1, 1), 1)
            distances_transform = distances.clone().div_(T).exp_()

            probs = torch.sum(
                torch.mul(
                    retrieval_one_hot.view(batch_size, -1, num_classes),
                    distances_transform.view(batch_size, -1, 1),
                ),
                1,
            )
            _, predictions = probs.sort(1, True)
            # predictions_np=predictions.cpu().numpy()

            # find the predictions that match the target
            correct = predictions.eq(targets.data.view(-1, 1))
            top1 = top1 + correct.narrow(1, 0, 1).sum().item()
            top5 = top5 + correct.narrow(1, 0, min(5, k)).sum().item()  # top5 does not make sense if k < 5
            total += targets.size(0)
        top1 = top1 * 100.0 / total
        top5 = top5 * 100.0 / total
        return top1, top5

    def imgs_align_v2(self, val_particles_path, val_particles_ctf_path, align_batch=10000, iteration_times=3,
                      multiprocessing_for_classes=True, resize=None, low_pass=None, pixel_size=None):

        k = self.k
        phbar = tqdm(range(k))
        phbar.set_description("rotating correction")
        average_imgs_path = self.result_path + "/averages/"

        raw_imgs_path_list = np.asarray(val_particles_path)
        raw_imgs_path_list_ctf = np.asarray(val_particles_ctf_path)

        labels = self.labels_p
        # pionts_nearest_to_centers = get_points_nearest_to_centers(self.features, self.centers, self.labels_p)
        for j in phbar:
            # find children and generate averages

            children_path = raw_imgs_path_list[labels == j]
            children_path = children_path[0:10000] if children_path.shape[0] > 10000 else children_path
            children_path_ctf = raw_imgs_path_list_ctf[labels == j]
            children_path_ctf = children_path_ctf[0:10000] if children_path_ctf.shape[0] > 10000 else children_path_ctf

            children = read_imgs_from_path_list_to_em(children_path)
            children_ctf = read_imgs_from_path_list_to_em(children_path_ctf)

            # ref = EMNumPy.numpy2em(np.asarray(pickle.load(open(raw_imgs_path_list[int(pionts_nearest_to_centers[j])], 'rb'))))

            # average_ctf = class_average(children_ctf,ref=ref)
            average_ctf = class_average(children_ctf)
            # average_ctf_np = to_int8(EMNumPy.em2numpy(average_ctf[0]))

            average = class_average_withali(images=children, ptcl_info=average_ctf[1], ref=average_ctf[0],
                                            xform=Transform(), focused=None)
            # from copy import deepcopy
            # average_np_1 = deepcopy(average)

            average_np = EMNumPy.em2numpy(average).copy()
            average_np = np.expand_dims(average_np, axis=0)
            # save the averages
            if j == 0:
                averages = average_np
            else:
                averages = np.append(averages, average_np, axis=0)
        if not os.path.exists(average_imgs_path):
            os.makedirs(average_imgs_path)
        # save_image(torch.unsqueeze(torch.from_numpy(averages), 1)
        #            average_imgs_path + '/clustering_result' + str(epoch) + '.png')
        # averages=mrcs_to_int8(averages)
        self.original_averages = averages
        averages = mrcs_norm_to_0_1(averages.copy())
        averages = np.expand_dims(averages, axis=1)
        self.averages = averages
        # average = np.expand_dims(average, axis=0)
        return averages

    def imgs_align_v2_multi(self, val_particles_path, val_particles_ctf_path, align_batch=10000, iteration_times=3,
                            multiprocessing_for_classes=True, resize=None, low_pass=None, pixel_size=None):

        average_imgs_path = self.result_path + "/averages/"
        if not os.path.exists(average_imgs_path):
            os.makedirs(average_imgs_path)

        averages_list = imgs_align_multi(k=self.k, labels=self.labels_p, val_particles_path=val_particles_path,
                                         val_particles_ctf_path=val_particles_ctf_path, resize=resize)

        # save_image(torch.unsqueeze(torch.from_numpy(averages), 1)
        #            average_imgs_path + '/clustering_result' + str(epoch) + '.png')
        # averages=mrcs_to_int8(averages)
        self.original_averages = averages_list
        averages = mrcs_norm_to_0_1(averages_list.copy())
        # averages = np.expand_dims(averages, axis=1)
        self.averages = averages
        # average = np.expand_dims(average, axis=0)
        return averages

    def imgs_align(self, val_particles_path=None, align_batch=10000, iteration_times=3,
                   multiprocessing_for_classes=True, resize=None, low_pass=None, pixel_size=None):
        k = self.k
        phbar = tqdm(range(k))
        phbar.set_description("rotating correction")
        average_imgs_path = self.result_path + "/averages/"
        total_num = 0
        path_raw_to_align = []
        if val_particles_path is not None:
            raw_imgs_path_list = np.asarray(val_particles_path)
        else:
            raw_imgs_path_list = np.asarray(self.particles_path)
        labels = self.labels_p
        pionts_nearest_to_centers = get_points_nearest_to_centers(self.features, self.centers, self.labels_p)
        for j in phbar:
            # find children and generate averages
            if isinstance(raw_imgs_path_list[0], str):
                children_path = raw_imgs_path_list[labels == j]
                path_raw_to_align += children_path.tolist()

                children_iter_times = int(children_path.shape[0] / align_batch)
                if children_path.shape[0] % align_batch > 0:
                    children_iter_times += 1
                # children_path=children_path[0:10000] if children_path.shape[0]>10000 else children_path
                children = read_imgs_from_path_list(children_path)
                if low_pass is not None and low_pass['use_low_pass']:
                    low_pass_scale = low_pass['scale']
                    children = mrcs_low_pass(children, low_pass_scale, pixel_size)
                if resize is not None:
                    children = mrcs_resize(children, resize, resize)
                # ref = EMNumPy.numpy2em(np.asarray(Image.open(raw_imgs_path_list[int(pionts_nearest_to_centers[j])])))
                ref = EMNumPy.numpy2em(
                    np.asarray(pickle.load(open(raw_imgs_path_list[int(pionts_nearest_to_centers[j])], 'rb'))))
            else:
                children = raw_imgs_path_list[labels == j]
                children = children[0:10000] if children.shape[0] > 10000 else children
                if low_pass is not None and low_pass['use_low_pass']:
                    children = mrcs_low_pass(children, low_pass, pixel_size)
                if resize is not None:
                    children = mrcs_resize(children, resize, resize)
                children_iter_times = int(children.shape[0] / align_batch)
                if children.shape[0] % align_batch > 0:
                    children_iter_times += 1
                ref = EMNumPy.numpy2em(raw_imgs_path_list[int(pionts_nearest_to_centers[j])])
            # children = children[np.asarray(reliable_list)[labels == j] == True]
            num_children = children.shape[0]
            for i in range(iteration_times):
                # print("Iter ", i)
                avgr = Averagers.get("mean", {"ignore0": True})
                average_list_i = []
                aligned_list = []
                for ii in range(children_iter_times):
                    # e=(ii+1)*align_batch if (ii+1)*align_batch<num_children else num_children+1
                    if (ii + 1) * align_batch < num_children:
                        item = [children[ii * align_batch:(ii + 1) * align_batch][i] for i in range(align_batch)]
                    else:
                        item = [children[ii * align_batch:][i] for i in range(num_children % align_batch)]
                    # item = [children for i in range(num_children)]
                    func = partial(aligin_one_img, ref=ref)
                    pool = multiprocessing.Pool(15)
                    aligined_img = pool.map(func, item)
                    aligned_list.append(aligined_img)
                    pool.close()
                    pool.join()
                    for num in range(len(aligined_img)):
                        avgr.add_image(aligined_img[num])
                        # aligined_imgs[total_num + num] = EMNumPy.em2numpy(aligined_img[num])
                    ref = avgr.finish()
            # avfr_np= EMNumPy.em2numpy(avgr)
            average = EMNumPy.em2numpy(ref)
            average = np.expand_dims(average, axis=0)
            # save the averages
            if j == 0:
                averages = average.copy()
            else:
                averages = np.append(averages, average, axis=0)
            total_num += num_children
        if not os.path.exists(average_imgs_path):
            os.makedirs(average_imgs_path)
        # save_image(torch.unsqueeze(torch.from_numpy(averages), 1),
        #            average_imgs_path + '/clustering_result' + str(epoch) + '.png')
        # averages=mrcs_to_int8(averages)
        self.original_averages = averages
        averages = mrcs_norm_to_0_1(averages.copy())
        averages = np.expand_dims(averages, axis=1)
        self.averages = averages
        # average = np.expand_dims(average, axis=0)
        return averages


def imgs_align_multi(k, labels, val_particles_path, val_particles_ctf_path, resize=None):
    phbar = tqdm(range(k))
    phbar.set_description("rotating correction")
    raw_imgs_path_list = np.asarray(val_particles_path)
    raw_imgs_path_list_ctf = np.asarray(val_particles_ctf_path)
    func = partial(class_align, raw_imgs_path_list=raw_imgs_path_list, raw_imgs_path_list_ctf=raw_imgs_path_list_ctf,
                   labels=labels, resize=resize)
    pool = multiprocessing.Pool(15)
    averages_list = np.asarray(pool.map(func, phbar))
    pool.close()
    pool.join()
    # averages = mrcs_norm_to_0_1(averages_list.copy())
    return averages_list


def my_save_averages(epoch, result_path, averages, original_averages):
    averages_predict_path = result_path + '/clustering_predict/epoch_' + str(epoch)
    if not os.path.exists(averages_predict_path):
        os.makedirs(averages_predict_path)
    with open(averages_predict_path + '/averages.data', 'wb') as filehandle:
        pickle.dump(original_averages, filehandle)
    if np.ndim(averages) == 3:
        averages = np.expand_dims(averages, axis=1)
    save_image(torch.tensor(averages), averages_predict_path + '/averages.png')


def best_map(L1, L2):
    # L1 should be the labels and L2 should be the downstream_clustering number we got
    Label1 = np.unique(L1)  # 去除重复的元素，由小大大排列
    nClass1 = len(Label1)  # 标签的大小
    Label2 = np.unique(L2)
    nClass2 = len(Label2)
    nClass = np.maximum(nClass1, nClass2)
    G = np.zeros((nClass, nClass))
    for i in range(nClass1):
        ind_cla1 = L1 == Label1[i]
        ind_cla1 = ind_cla1.astype(float)
        for j in range(nClass2):
            ind_cla2 = L2 == Label2[j]
            ind_cla2 = ind_cla2.astype(float)
            G[i, j] = np.sum(ind_cla2 * ind_cla1)
    m = Munkres()
    index = m.compute(-G.T)
    index = np.array(index)
    c = index[:, 1]
    newL2 = np.zeros(L2.shape)
    for i in range(nClass2):
        newL2[L2 == Label2[i]] = Label1[c[i]]
    return newL2


def get_points_nearest_to_centers(points_data, centers, labels):
    k = centers.shape[0]
    points_nearest_to_centers = np.zeros(k)
    mindist = np.full(k, np.inf)
    for i in range(points_data.shape[0]):
        class_id = labels[i]
        dist = np.linalg.norm(points_data[i] - centers[class_id])
        if dist < mindist[class_id]:
            points_nearest_to_centers[class_id] = i
            mindist[class_id] = dist
    return points_nearest_to_centers


def delete_file(filePath):
    if os.path.exists(filePath):
        for fileList in os.walk(filePath):
            for name in fileList[2]:
                os.chmod(os.path.join(fileList[0], name), stat.S_IWRITE)
                os.remove(os.path.join(fileList[0], name))
        shutil.rmtree(filePath)
        return "delete ok"
    else:
        return "no filepath"


def class_align(j, raw_imgs_path_list, raw_imgs_path_list_ctf, labels, resize):
    # find children and generate averages

    children_path = raw_imgs_path_list[labels == j]
    children_path = children_path[0:10000] if children_path.shape[0] > 10000 else children_path
    # children_path_ctf = raw_imgs_path_list_ctf[labels == j]
    children_path_ctf = raw_imgs_path_list[labels == j]
    children_path_ctf = children_path_ctf[0:10000] if children_path_ctf.shape[0] > 10000 else children_path_ctf

    children = read_imgs_from_path_list_to_em(children_path, resize=resize)

    children_ctf = read_imgs_from_path_list_to_em(children_path_ctf, resize=resize)

    # ref = EMNumPy.numpy2em(np.asarray(pickle.load(open(raw_imgs_path_list[int(pionts_nearest_to_centers[j])], 'rb'))))

    # average_ctf = class_average(children_ctf,ref=ref)
    average_ctf = class_average(children_ctf)
    # average_ctf_np = to_int8(EMNumPy.em2numpy(average_ctf[0]))

    average = class_average_withali(images=children, ptcl_info=average_ctf[1], ref=average_ctf[0],
                                    xform=Transform(), focused=None)
    # from copy import deepcopy
    # average_np_1 = deepcopy(average)

    average_np = EMNumPy.em2numpy(average).copy()
    average_np = np.expand_dims(average_np, axis=0)

    return average_np


def aligin_one_img(im, ref):
    im = EMNumPy.numpy2em(im)
    # im.process_inplace("normalize.edgemean")
    # if im["nx"]!=nx or im["ny"]!=ny :
    # 	im=im.get_clip(Region(old_div(-(nx-im["nx"]),2),old_div(-(ny-im["ny"]),2),nx,ny))
    # im.write_image("result/seq.mrc",-1)
    # ima=im.align("translational",ref0,{"nozero":1,"maxshift":old_div(ref0["nx"],4.0)},"ccc",{})
    ima = im.align("rotate_translate_tree", ref)
    # ima = EMNumPy.em2numpy(ima)
    return ima


def read_imgs_from_path_list(paths):
    imgs = []
    for path in paths:
        # imgs.append(np.asarray(Image.open(path)))
        imgs.append(np.asarray(pickle.load(open(path, 'rb'))))
    return np.asarray(imgs)


def read_imgs_from_path_list_to_em(paths, resize):
    imgs = []
    for path in paths:
        if resize is not None:
            img = pickle.load(open(path, 'rb')).resize((resize, resize), Image.Resampling.BICUBIC)
        else:
            img = pickle.load(open(path, 'rb'))
        # imgs.append(np.asarray(Image.open(path)))
        imgs.append(EMNumPy.numpy2em(np.asarray(img)))
    return imgs


def mrcs_norm_to_0_1(mrcs):
    if torch.is_tensor(mrcs):
        for i in range(mrcs.shape[0]):
            # m=torch.max(mrcs[i])
            mrcs[i] = (mrcs[i] - torch.min(mrcs[i])) / (torch.max(mrcs[i]) - torch.min(mrcs[i]))
    else:
        for i in range(mrcs.shape[0]):
            mrcs[i] = (mrcs[i] - np.min(mrcs[i])) / ((np.max(mrcs[i]) - np.min(mrcs[i])))
    return mrcs


def mrcs_low_pass(mrcs, cutoff_A, pixelsize):
    img_width = mrcs.shape[1]
    cutoff_pixel = int(pixelsize * img_width / cutoff_A)
    mask = filter(img_width, cutoff_pixel, filter='gaussian', type='lp')
    for i in range(mrcs.shape[0]):
        img = mrcs[i]
        f = cv2.dft(img.astype(np.float32), flags=cv2.DFT_COMPLEX_OUTPUT)
        f_shifted = np.fft.fftshift(f)

        # f_filtered = f_shifted * mask
        f_complex = f_shifted[:, :, 0] * 1j + f_shifted[:, :, 1]
        f_filtered = mask * f_complex
        f_filtered_shifted = np.fft.fftshift(f_filtered)
        inv_img = np.fft.ifft2(f_filtered_shifted)
        filtered_img = np.real(inv_img)
        mrcs[i] = filtered_img
    return mrcs


def get_averages_with_labels():
    # from e2classaverage import class_average_withali,class_average
    # from EMAN2 import Transform
    # label_path='/yanyang2/dataset/clustering/cs_simulated_particles/emd_6840/withlabels/valset_snr01_5/output_tif_label.data'
    # label_path='/yanyang2/projects/results/particle_clustering_inference/emd6840_valset/1220_snr01/test1/clustering_predict/epoch_291/output_tif_label.data'
    label_path = '/yanyang2/dataset/clustering/cs_simulated_particles/j1461/withlabels/realdata_1228_ns05/cs_predict_label.data'
    # imgs_path='/yanyang2/dataset/clustering/cs_simulated_particles/emd_6840/withlabels/valset_snr01_5/output_tif_path.data'
    imgs_path = '/yanyang2/dataset/clustering/cs_simulated_particles/j1461/withlabels/realdata_1228_ns05/output_tif_path.data'
    # imgs_path_ctf='/yanyang2/dataset/clustering/cs_simulated_particles/emd_6840/withlabels/valset_snr01_5/output_ctf_tif_path.data'
    imgs_path_ctf = '/yanyang2/dataset/clustering/cs_simulated_particles/j1461/withlabels/realdata_1228_ns05/output_ctf_tif_path.data'
    k = 9
    save_path = '/yanyang2/dataset/test/average_test/'
    with open(label_path, 'rb') as filehandle:
        label_path_list = np.asarray(pickle.load(filehandle))
    with open(imgs_path, 'rb') as filehandle:
        imgs_path_list = np.asarray(pickle.load(filehandle))
    with open(imgs_path_ctf, 'rb') as filehandle:
        imgs_path_ctf_list = np.asarray(pickle.load(filehandle))

    if not os.path.exists(save_path):
        os.makedirs(save_path)

    averages_list = imgs_align_multi(k=k, labels=label_path_list, val_particles_path=imgs_path_list,
                                     val_particles_ctf_path=imgs_path_ctf_list, resize=128)
    averages = mrcs_norm_to_0_1(averages_list.copy())
    my_save_averages(0, save_path, averages, averages_list)

    # em_imgs_list=[]
    # em_imgs_list_ctf=[]
    # children_path = imgs_path_list[label_path_list == 2]
    # children_path_ctf = imgs_path_ctf_list[label_path_list == 2]
    # children=read_imgs_from_path_list(children_path)
    # # children_ctf=read_imgs_from_path_list(children_path)
    # children_ctf=read_imgs_from_path_list(children_path_ctf)
    # for i in range(children.shape[0]):
    #     em_imgs_list.append(EMNumPy.numpy2em(children[i]))
    #     em_imgs_list_ctf.append(EMNumPy.numpy2em(children_ctf[i]))
    # # children_em=EMNumPy.numpy2em(children)
    # # average=class_average_withali(children_em,iteration_times=3)
    # average=class_average(em_imgs_list)
    # average_np=to_int8(EMNumPy.em2numpy(average[0]))
    # Image.fromarray(average_np).save(save_path+'average.tif')
    # average_2=class_average_withali(images=em_imgs_list_ctf,ptcl_info=average[1],ref=average[0],xform= Transform(),focused=None)
    # average_np_ctf=to_int8(EMNumPy.em2numpy(average_2))
    # Image.fromarray(average_np_ctf).save(save_path+'average2.tif')


def save_augmented_imgs(
        augmented_imgs: torch.Tensor,
        weights: torch.Tensor,
        save_path: str,
        weight_bar: float = 0.0,
        img_name_prefix: str = 'augmented_img',
        images_per_row: int = 10,
        font_color: tuple = (255, 0, 0),  # 醒目的红色
        sorted_by_weight: bool = True,
):
    from PIL import Image, ImageDraw, ImageFont
    """
    筛选权重达标的图像，将权重值标注在图像上，然后将它们拼接成一张大图并保存。

    Args:
        augmented_imgs (torch.Tensor): 尺寸为 (N, 1, dim, dim) 的图像张量。
        weights (torch.Tensor): 尺寸为 (N,) 的权重张量，数值范围 0~1。
        save_path (str): 图像保存的目录路径。
        weight_bar (float, optional): 权重的筛选阈值。默认为 0.1。
        img_name_prefix (str, optional): 保存图像文件的前缀名。默认为 'augmented_img'。
        images_per_row (int, optional): 网格布局中每行的图像数量。默认为 10。
        font_color (tuple, optional): 标注文字的RGB颜色。默认为红色 (255, 0, 0)。
    """
    # 检查输入尺寸是否匹配
    if augmented_imgs.shape[0] != weights.shape[0]:
        raise ValueError(f"图像数量 ({augmented_imgs.shape[0]}) 和权重数量 ({weights.shape[0]}) 不匹配。")

    # 1. 按照 weight_bar 筛选图像和权重
    # 创建布尔掩码
    mask = weights > weight_bar
    filtered_imgs = augmented_imgs[mask]
    filtered_weights = weights[mask]

    num_filtered_imgs = filtered_imgs.shape[0]

    if num_filtered_imgs == 0:
        print(f"没有图像的权重超过 {weight_bar}，不保存任何图像。")
        return

    print(f"共有 {num_filtered_imgs} 张图像的权重超过 {weight_bar}，将进行处理。")

    # 2. 将权重标注在图像上
    annotated_imgs = []
    # 获取图像尺寸
    _, _, dim, _ = filtered_imgs.shape
    imgs_np=mrcs_to_int8(filtered_imgs.squeeze().cpu().numpy())  # 转换为 uint8 格式的 numpy 数组

    if sorted_by_weight:
        # 按照权重从大到小排序
        sorted_indices = torch.argsort(filtered_weights, descending=True)
    else:
        # 按照原始顺序
        sorted_indices = torch.arange(num_filtered_imgs)


    for i in sorted_indices:
        # img_tensor = filtered_imgs[i]  # Shape: (1, dim, dim)
        weight_value = filtered_weights[i].item()

        # 将单通道灰度图转换为 PIL Image 对象 (需要先转为 numpy)
        # 乘以 255 并转换为 uint8 类型
        # img_np = img_tensor.squeeze().cpu().numpy() * 255
        # pil_img = Image.fromarray(img_np.astype(np.uint8), 'L')
        pil_img =Image.fromarray(imgs_np[i].astype(np.uint8), 'L')
        # 转换为 RGB 模式以便添加彩色文字
        pil_img_rgb = pil_img.convert('RGB')

        # 在图像上绘制文字
        draw = ImageDraw.Draw(pil_img_rgb)
        text = f"{weight_value:.3f}"  # 格式化权重值为3位小数

        # 定义文字位置（左上角，留出一些边距）
        text_position = (5, 5)

        # 绘制文字
        draw.text(text_position, text, fill=font_color)

        # 将 PIL Image 对象转换回 Tensor
        # PIL (H, W, C) -> Numpy -> Tensor (C, H, W)
        back_to_np = np.array(pil_img_rgb)
        back_to_tensor = torch.from_numpy(back_to_np).permute(2, 0, 1)  # / 255.0 if normalization is needed
        annotated_imgs.append(back_to_tensor)

    if not annotated_imgs:
        print("没有可供处理的图像。")
        return

    # 将标注好的图像列表堆叠成一个 batch
    annotated_batch = torch.stack(annotated_imgs)

    # 3. 检查并创建保存目录
    os.makedirs(save_path, exist_ok=True)

    # 4. 组合图像并保存
    save_file_path = os.path.join(save_path, f"{img_name_prefix}.png")

    # 使用 torchvision.utils.save_image 直接生成网格图并保存
    save_image(
        annotated_batch.float() / 255.0,  # save_image 需要 [0,1] 范围的浮点数张量
        save_file_path,
        nrow=images_per_row,
        normalize=False  # 因为我们已经处理好了范围，不需要它再标准化
    )
    # print(f"图像已成功保存至: {save_file_path}")


def get_imgs_by_ids(dataloader, target_ids, pose_items_id,min_id=0,name='aug1'):
    dataset = dataloader.dataset
    results = []

    for idx in target_ids:
        results.append(dataset[pose_items_id[idx]+min_id][name].cpu().numpy())

    results =torch.from_numpy( np.asarray(results))  # 转换为形状 (N, 1, dim, dim)
    return results

if __name__ == '__main__':
    get_averages_with_labels()



import os
import struct
from collections import OrderedDict

import numpy as np

# See ref:
# MRC2014: Extensions to the MRC format header for electron cryo-microscopy and tomography
# And:
# https://www.ccpem.ac.uk/mrc_format/mrc2014.php

DTYPE_FOR_MODE = {0: np.int8,
                  1: np.int16,
                  2: np.float32,
                  3: '2h',  # complex number from 2 shorts
                  4: np.complex64,
                  6: np.uint16,
                  12: np.float16,
                  16: '3B'}  # RBG values
MODE_FOR_DTYPE = {vv: kk for kk, vv in DTYPE_FOR_MODE.items()}


class MRCHeader:
    """MRC header class"""
    FIELDS = ['nx', 'ny', 'nz',  # int
              'mode',  # int
              'nxstart', 'nystart', 'nzstart',  # int
              'mx', 'my', 'mz',  # int
              'xlen', 'ylen', 'zlen',  # float
              'alpha', 'beta', 'gamma',  # float
              'mapc', 'mapr', 'maps',  # int
              'amin', 'amax', 'amean',  # float
              'ispg', 'next', 'creatid',  # int, int, short, [pad 30]
              'nint', 'nreal',  # short, [pad 20]
              'imodStamp', 'imodFlags',  # int
              'idtype', 'lens', 'nd1', 'nd2', 'vd1', 'vd2',  # short
              'tilt_ox', 'tilt_oy', 'tilt_oz',  # float
              'tilt_cx', 'tilt_cy', 'tilt_cz',  # float
              'xorg', 'yorg', 'zorg',  # float
              'cmap', 'stamp', 'rms',  # char[4], float
              'nlabl', 'labels']  # int, char[10][80]
    FSTR = '3ii3i3i3f3f3i3f2ih30x2h20x2i6h6f3f4s4sfi800s'
    STRUCT = struct.Struct(FSTR)

    def __init__(self, header_values, extended_header=b''):
        self.fields = OrderedDict(zip(self.FIELDS, header_values))
        self.extended_header = extended_header
        self.D = self.fields['nx']
        self.dtype = DTYPE_FOR_MODE[self.fields['mode']]

    def __str__(self):
        return f'Header: {self.fields}\nExtended header: {self.extended_header}'

    @classmethod
    def parse(cls, fname):
        with open(fname, 'rb') as f:
            header = cls(cls.STRUCT.unpack(f.read(1024)))
            extbytes = header.fields['next']
            extended_header = f.read(extbytes)
            header.extended_header = extended_header
        return header

    @classmethod
    def make_default_header(cls, data, is_vol=True, a_pix=1., xorg=0., yorg=0., zorg=0.):
        nz, ny, nx = data.shape
        ispg = 1 if is_vol else 0
        if is_vol:
            dmin, dmax, dmean, rms = data.min(), data.max(), data.mean(), data.std()
        else:  # use undefined values for image stacks
            dmin, dmax, dmean, rms = -1, -2, -3, -1
        vals = [nx, ny, nz,
                2,  # mode = 2 for 32-bit float
                0, 0, 0,  # nxstart, nystart, nzstart
                nx, ny, nz,  # mx, my, mz
                a_pix * nx, a_pix * ny, a_pix * nz,  # cella
                90., 90., 90.,  # cellb
                1, 2, 3,  # mapc, mapr, maps
                dmin, dmax, dmean,
                ispg,
                0,  # exthd_size
                0,  # creatid
                0, 0,  # nint, nreal
                0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0,
                xorg, yorg, zorg,
                b'MAP ' if is_vol else b'\x00' * 4, b'\x00' * 4,  # cmap, stamp
                rms,  # rms
                0,  # nlabl
                b'\x00' * 800,  # labels
                ]
        return cls(vals)

    def write(self, fh):
        buf = self.STRUCT.pack(*list(self.fields.values()))
        fh.write(buf)
        fh.write(self.extended_header)

    def get_apix(self):
        return self.fields['xlen'] / self.fields['nx']

    def update_apix(self, a_pix):
        self.fields['xlen'] = self.fields['nx'] * a_pix
        self.fields['ylen'] = self.fields['ny'] * a_pix
        self.fields['zlen'] = self.fields['nz'] * a_pix

    def get_origin(self):
        return self.fields['xorg'], self.fields['yorg'], self.fields['zorg']

    def update_origin(self, xorg, yorg, zorg):
        self.fields['xorg'] = xorg
        self.fields['yorg'] = yorg
        self.fields['zorg'] = zorg


class LazyImage:
    """On-the-fly image loading"""

    def __init__(self, fname, shape, dtype, offset):
        self.fname = fname
        self.shape = shape
        self.dtype = dtype
        self.offset = offset

    def get(self):
        with open(self.fname) as f:
            f.seek(self.offset)
            image = np.fromfile(f, dtype=self.dtype, count=np.prod(self.shape)).reshape(self.shape)
        return image


def parse_header(fname):
    return MRCHeader.parse(fname)


def parse_mrc_list(txtfile, lazy=False):
    lines = open(txtfile, 'r').readlines()

    def abspath(f):
        if os.path.isabs(f):
            return f
        base = os.path.dirname(os.path.abspath(txtfile))
        return os.path.join(base, f)

    lines = [abspath(x) for x in lines]
    if not lazy:
        particles = np.vstack([parse_mrc(x.strip(), lazy=False)[0] for x in lines])
    else:
        particles = [img for x in lines for img in parse_mrc(x.strip(), lazy=True)[0]]
    return particles


def parse_mrc(fname, lazy=False):
    # parse the header
    header = MRCHeader.parse(fname)

    # get the number of bytes in extended header
    extbytes = header.fields['next']
    start = 1024 + extbytes  # start of image data

    dtype = header.dtype
    nz, ny, nx = header.fields['nz'], header.fields['ny'], header.fields['nx']

    # load all in one block
    if not lazy:
        with open(fname, 'rb') as fh:
            fh.read(start)  # skip the header + extended header
            array = np.fromfile(fh, dtype=dtype).reshape((nz, ny, nx))

    # or list of LazyImages
    else:
        stride = dtype().itemsize * ny * nx
        array = [LazyImage(fname, (ny, nx), dtype, start + i * stride) for i in range(nz)]
    return array, header


def write(fname, array, header=None, a_pix=1., xorg=0., yorg=0., zorg=0., is_vol=None):
    # get a default header
    if header is None:
        if is_vol is None:
            is_vol = True if len(set(array.shape)) == 1 else False  # Guess whether data is vol or image stack
        header = MRCHeader.make_default_header(array, is_vol, a_pix, xorg, yorg, zorg)
    # write the header
    f = open(fname, 'wb')
    header.write(f)
    f.write(array.tobytes())



from __future__ import print_function
import numpy as np
from .pyem import metadata
from .pyem import star
def cs2star(input, output,minphic=None,boxsize=None,noswapxy=False,invertx=False,inverty=False):
    # parser = argparse.ArgumentParser()
    # # parser.add_argument("input", help="Cryosparc metadata .csv (v0.6.5) or .cs (v2+) files", nargs="*")
    # # parser.add_argument("output", help="Output .star file")
    # parser.add_argument("--movies", help="Write per-movie star files into output directory", action="store_true")
    # parser.add_argument("--boxsize", help="Cryosparc refinement box size (if different from particles)", type=float)
    # # parser.add_argument("--passthrough", "-p", help="List file required for some Cryosparc 2+ job types")
    # parser.add_argument("--class", help="Keep this class in output, may be passed multiple times",
    #                     action="append", type=int, dest="cls")
    # parser.add_argument("--minphic", help="Minimum posterior probability for class assignment", type=float, default=0)
    # parser.add_argument("--stack-path", help="Path to single particle stack", type=str)
    # parser.add_argument("--micrograph-path", help="Replacement path for micrographs or movies")
    # parser.add_argument("--copy-micrograph-coordinates",
    #                     help="Source for micrograph paths and particle coordinates (file or quoted glob)",
    #                     type=str)
    # parser.add_argument("--swapxy",
    #                     help="Swap X and Y axes when converting particle coordinates from normalized to absolute",
    #                     action="store_true")
    # parser.add_argument("--noswapxy", help="Do not swap X and Y axes when converting particle coordinates",
    #                     action="store_false")
    # parser.add_argument("--invertx", help="Invert particle coordinate X axis", action="store_true")
    # parser.add_argument("--inverty", help="Invert particle coordinate Y axis", action="store_false")
    # parser.add_argument("--flipy", help="Invert refined particle Y shifts", action="store_true")
    # parser.add_argument("--cached", help="Keep paths from the Cryosparc 2+ cache when merging coordinates",
    #                     action="store_true")
    # parser.add_argument("--transform",
    #                     help="Apply rotation matrix or 3x4 rotation plus translation matrix to particles (Numpy format)",
    #                     type=str)
    # parser.add_argument("--relion2", "-r2", help="Relion 2 compatible outputs", action="store_true")
    # parser.add_argument("--strip-uid", help="Strip all leading UIDs from file names", nargs="?", default=None, const=-1,
    #                     type=int)
    # parser.add_argument("--10k", help="Only read first 10,000 particles for rapid testing.", action="store_true",
    #                     dest="first10k")
    # parser.add_argument("--loglevel", "-l", type=str, default="WARNING", help="Logging level and debug output")

    # args=parser.parse_args()
    if isinstance(input,str):
        input=[input,]
    else:
        input=input
    # args.output=output

    # log = logging.getLogger('root')
    # hdlr = logging.StreamHandler(sys.stdout)
    # log.addHandler(hdlr)
    # log.setLevel(logging.getLevelName(args.loglevel.upper()))

    # if args.swapxy:
    #     log.warning("Axis swapping is now the default and --swapxy has no effect. "
    #                 "Use --noswapxy if unswapping is needed (unlikely).")


    if input[0].endswith(".cs"):
        # log.debug("Detected CryoSPARC 2+ .cs file")
        cs = np.load(input[0])
        # if first10k:
        #     cs = cs[:10000]
        #
        # if movies:
        #     if not os.path.isdir(output):
        #         # log.error("%s is not a directory" % args.output)
        #         return 1
        #     # log.info("Writing per-movie star files into %s" % args.output)
        #     trajdir = os.path.dirname(os.path.dirname(input[0]))
        #     if len(input) > 1:
        #         pts = [a for a in input[1:] if a.endswith(".cs")]
        #     else:
        #         pts = None
        #     data_general = metadata.cryosparc_2_cs_movie_parameters(cs, passthroughs=pts, trajdir=trajdir, path=micrograph_path)
        #     data_general[star.Relion.MICROGRAPHMETADATA] = data_general[star.Relion.MICROGRAPH_NAME].apply(
        #         lambda x: os.path.join(output, os.path.basename(x.rstrip(".mrc")) + ".star"))
        #     for mic in metadata.cryosparc_2_cs_motion_parameters(cs, data_general, trajdir=trajdir):
        #         fn = mic[star.Relion.GENERALDATA][star.Relion.MICROGRAPHMETADATA]
        #         # log.debug("Writing %s" % fn)
        #         star.write_star_tables(fn, mic)
        #     fields = [star.Relion.VOLTAGE, star.Relion.CS, star.Relion.AC, star.Relion.MICROGRAPHORIGINALPIXELSIZE,
        #               star.Relion.MICROGRAPHPIXELSIZE, star.Relion.MICROGRAPH_NAME, star.Relion.MICROGRAPHMETADATA,
        #               star.Relion.MICROGRAPHBINNING, star.Relion.OPTICSGROUP]
        #     if len(input) > 1 and input[-1].endswith(".star"):
        #         mic_star = input[-1]
        #         star.write_star(mic_star, data_general[[f for f in fields if f in data_general]])
        #     return 0

        try:
            df = metadata.parse_cryosparc_2_cs(cs, passthroughs=input[1:], minphic=minphic,
                                               boxsize=boxsize, swapxy=noswapxy,
                                               invertx=invertx, inverty=inverty)
        except (KeyError, ValueError) as e:
            # log.error(e, exc_info=True)
            # log.error("Required fields could not be mapped. Are you using the right input file(s)?")
            return 1
    else:
        # log.debug("Detected CryoSPARC 0.6.5 .csv file")
        if len(input) > 1:
            # log.error("Only one file at a time supported for CryoSPARC 0.6.5 .csv format")
            return 1
        meta = metadata.parse_cryosparc_065_csv(input[0])  # Read cryosparc metadata file.
        df = metadata.cryosparc_065_csv2star(meta, minphic)

    # if cls is not None:
    #     df = star.select_classes(df, cls)
    #
    # if flipy:
    #     # log.info("Flipping refined shifts in Y")
    #     df[star.Relion.ORIGINY] = -df[star.Relion.ORIGINY]
    #     # log.info("Flipping particle orientation through XZ plane")
    #     df = star.transform_star(df, np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]]), leftmult=True)
    #
    # if strip_uid is not None:
    #     df = star.strip_path_uids(df, inplace=True, count=args.strip_uid)
    #
    # if copy_micrograph_coordinates is not None:
    #     df = star.augment_star_ucsf(df, inplace=True)
    #     coord_star = pd.concat(
    #         (star.parse_star(inp, keep_index=False, augment=True) for inp in
    #          glob(copy_micrograph_coordinates)), join="inner")
    #     key = star.merge_key(df, coord_star, threshold=0)
    #     if key is None and not strip_uid:
    #         # log.debug("Merge key not found, removing leading UIDs")
    #         df = star.strip_path_uids(df, inplace=True)
    #         key = star.merge_key(df, coord_star)
    #     # log.debug("Coordinates merge key: %s" % key)
    #     if cached or key == star.Relion.IMAGE_NAME:
    #         fields = star.Relion.MICROGRAPH_COORDS
    #     else:
    #         fields = star.Relion.MICROGRAPH_COORDS + [star.UCSF.IMAGE_INDEX, star.UCSF.IMAGE_PATH]
    #     n = df.shape[0]
    #     df = star.smart_merge(df, coord_star, fields=fields, key=key)
    #     star.simplify_star_ucsf(df)
    #     # if df.shape[0] != n:
    #         # log.warning("%d / %d particles remain after coordinate merge" % (df.shape[0], n))

    # if micrograph_path is not None:
    #     df = star.replace_micrograph_path(df, micrograph_path, inplace=True)
    #
    # if transform is not None:
    #     r = np.array(json.loads(transform))
    #     df = star.transform_star(df, r, inplace=True)

    df = star.check_defaults(df, inplace=True)

    # if relion2:
    #     df = star.remove_new_relion31(df, inplace=True)
    #     star.write_star(output, df, resort_records=True, optics=False)
    # else:
    df = star.remove_deprecated_relion2(df, inplace=True)
    star.write_star(output, df, resort_records=True, optics=True)

    # log.info("Output fields: %s" % ", ".join(df.columns))
    # return df






# Copyright (C) 2018 Daniel Asarnow
# University of California, San Francisco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numba
import numpy as np


@numba.jit(cache=False, nopython=True, nogil=True)
def _bincount_nb(bins, data, out):
    for i in range(len(bins)):
        out[bins[i]] += data[i]
    return out


@numba.jit(cache=False, nopython=True, nogil=True)
def bincount_nb(bins, data, out=None):
    if out is None:
        out = np.zeros(bins.max(), dtype=data.dtype)
    _bincount_nb(bins, data, out)
    return out


@numba.jit(cache=False, nopython=True, nogil=True)
def bincorr_nb(p1, p2, bins, n=-1):
    bflat = bins.reshape(-1)
    p1flat = p1.reshape(-1)
    p2flat = p2.reshape(-1)
    if n == -1:
        n = np.max(bflat)
    fc = p1flat * np.conj(p2flat)
    fcc = _bincount_nb(bflat, fc, out=np.zeros(n, dtype=np.complex128))
    p1r = _bincount_nb(bflat, np.abs(p1flat)**2, out=np.zeros(n, dtype=np.float64))
    p2r = _bincount_nb(bflat, np.abs(p2flat)**2, out=np.zeros(n, dtype=np.float64))
    mag = np.sqrt(p1r * p2r)
    mag[-1] += 1e-17
    frc = fcc / mag
    return frc



# Copyright (C) 2017-2018 Daniel Asarnow
# University of California, San Francisco
#
# Library for interconversion of rotation representations.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numba
import numpy as np


@numba.jit(nopython=True, nogil=True)
def rot2euler(r, out=None):
    """Decompose rotation matrix into Euler angles"""
    # assert(isrotation(r))
    # Shoemake rotation matrix decomposition algorithm with same conventions as Relion.
    # epsilon = np.finfo(np.double).eps
    epsilon = 1e-16
    r = r.reshape(-1, 9).reshape(-1, 3, 3)
    if out is None:
        out = np.zeros((len(r), 3), dtype=r.dtype)
    for i in range(len(r)):
        abs_sb = np.sqrt(r[i, 0, 2] ** 2 + r[i, 1, 2] ** 2)
        if abs_sb > 16 * epsilon:
            gamma = np.arctan2(r[i, 1, 2], -r[i, 0, 2])
            alpha = np.arctan2(r[i, 2, 1], r[i, 2, 0])
            if np.abs(np.sin(gamma)) < epsilon:
                sign_sb = np.sign(-r[i, 0, 2]) / np.cos(gamma)
            else:
                sign_sb = np.sign(r[i, 1, 2]) if np.sin(gamma) > 0 else -np.sign(r[i, 1, 2])
            beta = np.arctan2(sign_sb * abs_sb, r[i, 2, 2])
        else:
            if np.sign(r[i, 2, 2]) > 0:
                alpha = 0
                beta = 0
                gamma = np.arctan2(-r[i, 1, 0], r[i, 0, 0])
            else:
                alpha = 0
                beta = np.pi
                gamma = np.arctan2(r[i, 1, 0], -r[i, 0, 0])
        out[i, 0] = alpha
        out[i, 1] = beta
        out[i, 2] = gamma
    return out


@numba.jit(nopython=True, nogil=True)
def euler2rot(alpha, beta, gamma):
    ca = np.cos(alpha)
    cb = np.cos(beta)
    cg = np.cos(gamma)
    sa = np.sin(alpha)
    sb = np.sin(beta)
    sg = np.sin(gamma)
    cc = cb * ca
    cs = cb * sa
    sc = sb * ca
    ss = sb * sa
    r = np.array(((cg * cc - sg * sa, cg * cs + sg * ca, -cg * sb),
                  (-sg * cc - cg * sa, -sg * cs + cg * ca, sg * sb),
                  (sc, ss, cb)))
    return r


@numba.jit(nopython=True, nogil=True)
def vec2rot(v):
    ax = v / np.linalg.norm(v)
    return euler2rot(*np.array([np.arctan2(ax[1], ax[0]), np.arccos(ax[2]), 0.]))


@numba.jit(nopython=True, nogil=True)
def quat2aa(q):
    n = np.sqrt(np.sum(q[1:] ** 2))
    theta = 2 * np.arctan2(np.abs(n), q[0])  # Or 2 * np.arccos(q[0])
    if np.abs(theta) > 1e-12:
        ax = q[1:] / np.sin(0.5 * theta)
    else:
        ax = np.zeros(3, dtype=q.dtype)
    return theta * ax


@numba.jit(nopython=True, nogil=True)
def aa2quat(ax):
    theta = np.linalg.norm(ax)
    if theta == 0:
        return np.array([1, 0, 0, 0], dtype=ax.dtype)
    q = np.zeros(4, dtype=ax.dtype)
    q[0] = np.cos(0.5 * theta)
    q[1:] = np.sin(0.5 * theta)
    q[1:] *= ax
    q[1:] /= theta
    if q[0] < 0:
        q[:] = -q[:]
    return q


@numba.jit(nopython=True, nogil=True)
def quat2rot(q):
    # aa = q[0] ** 2
    bb = 2 * q[1] ** 2
    cc = 2 * q[2] ** 2
    dd = 2 * q[3] ** 2
    ab = 2 * q[0] * q[1]
    ac = 2 * q[0] * q[2]
    ad = 2 * q[0] * q[3]
    bc = 2 * q[1] * q[2]
    bd = 2 * q[1] * q[3]
    cd = 2 * q[2] * q[3]
    # These formulas are equivalent forms taken from Wikipedia, but are transposed.
    # r = np.array([[aa + bb - cc - dd, 2*bc - 2*ad,       2*bd + 2*ac],
    #               [2*bc + 2*ad,       aa - bb + cc - dd, 2*cd - 2*ab],
    #               [2*bd - 2*ac,       2*cd + 2*ab,       aa - bb - cc + dd]], dtype=q.dtype)
    # r = np.array([[1 - 2 * (cc + dd), 2 * (bc - ad), 2 * (bd + ac)],
    #               [2 * (bc + ad), 1 - 2 * (bb + dd), 2 * (cd - ab)],
    #               [2 * (bd - ac), 2 * (cd + ab), 1 - 2 * (bb + cc)]], dtype=q.dtype)
    # Tentatively correct formula (assuming multiplication by 2):
    r = np.array([[1 - cc - dd, bc + ad, bd - ac],
                  [bc - ad, 1 - bb - dd, cd + ab],
                  [bd + ac, cd - ab, 1 - bb - cc]], dtype=q.dtype)
    # This formula is incorrect after fixing euler2quat for ZYZ.
    # r = np.array([[1 - 2 * (bb + dd), 2 * (ad - bc), -2 * (cd + ab)],
    #               [-2 * (bc + ad), 1 - 2 * (cc + dd), 2 * (bd - ac)],
    #               [2 * (ab - cd), 2 * (bd + ac), 1 - 2 * (bb + cc)]], dtype=q.dtype)
    # This assumes the multiplications by 2 are done first:
    # r = np.array([[1 - (bb + dd), (ad - bc), -(cd + ab)],
    #               [-(bc + ad), 1 - (cc + dd), (bd - ac)],
    #               [(ab - cd), (bd + ac), 1 - (bb + cc)]], dtype=q.dtype)
    return r


@numba.jit(nopython=True, nogil=True)
def rot2quat(r):
    q = np.zeros(4, dtype=r.dtype)
    tr = np.trace(r)
    if tr > 0:
        q[0] = np.sqrt(tr + 1) / 2
        sinv = 1 / (q[0] * 4)
        q[1] = sinv * (r[1, 2] - r[2, 1])
        q[2] = sinv * (r[2, 0] - r[0, 2])
        q[3] = sinv * (r[0, 1] - r[1, 0])
    else:
        mi = np.argmax(np.diag(r))
        if mi == 0:
            q[1] = np.sqrt(r[0, 0] - r[1, 1] - r[2, 2] + 1) / 2
            sinv = 1 / (q[1] * 4)
            q[0] = sinv * (r[1, 2] - r[2, 1])
            q[2] = sinv * (r[0, 1] + r[1, 0])
            q[3] = sinv * (r[0, 2] + r[2, 0])
        elif mi == 1:
            q[2] = np.sqrt(r[1, 1] - r[2, 2] - r[0, 0] + 1) / 2
            sinv = 1 / (q[2] * 4)
            q[0] = sinv * (r[2, 0] - r[0, 2])
            q[1] = sinv * (r[0, 1] + r[1, 0])
            q[3] = sinv * (r[1, 2] + r[2, 1])
        else:
            q[3] = np.sqrt(r[2, 2] - r[0, 0] - r[1, 1] + 1) / 2
            sinv = 1 / (q[3] * 4)
            q[0] = sinv * (r[0, 1] - r[1, 0])
            q[1] = sinv * (r[0, 2] + r[2, 0])
            q[2] = sinv * (r[1, 2] + r[2, 1])
    if q[0] < 0:
        q[:] = -q[:]
    return q


@numba.jit(nopython=True, nogil=True)
def euler2quat(alpha, beta, gamma):
    q = np.array([np.cos((alpha + gamma) / 2) * np.cos(beta / 2),
                  np.sin((gamma - alpha) / 2) * np.sin(beta / 2),
                  np.cos((alpha - gamma) / 2) * np.sin(beta / 2),
                  np.sin((alpha + gamma) / 2) * np.cos(beta / 2)])
    return q


@numba.jit(nopython=True, nogil=True)
def quat2euler(q):
    aa = q[0] ** 2
    bb = q[1] ** 2
    cc = q[2] ** 2
    dd = q[3] ** 2
    ab = q[0] * q[1]
    ac = q[0] * q[2]
    bd = q[1] * q[3]
    cd = q[2] * q[3]

    # alpha = np.arctan2(bd + ac, ab - cd)
    alpha = np.arctan2(np.abs(ab - cd), bd + ac)

    # if alpha < 0:
    #     alpha += 2 * np.pi

    beta = np.arccos(aa - bb - cc + dd)

    # gamma = np.arctan2(bd - ac, ab + cd)
    gamma = 2 * np.arctan2(np.abs(bd - ac), ab + cd)

    # if gamma < 0:
    #     gamma += 2 * np.pi

    return alpha, beta, gamma


@numba.jit(nopython=True, nogil=True)
def expmap(e, out=None):
    """Convert axis-angle vector into 3D rotation matrix"""
    # theta = np.linalg.norm(e)
    # if theta < 1e-16:
    #     return np.identity(3, e.dtype)
    # w = e / theta
    # k = np.array([[0, w[2], -w[1]],
    #               [-w[2], 0, w[0]],
    #               [w[1], -w[0], 0]], dtype=e.dtype)
    # r = np.identity(3, e.dtype) + np.sin(theta) * k + (np.ones(1, dtype=e.dtype) - np.cos(theta)) * np.dot(k, k)
    e = np.atleast_2d(e)
    if out is None:
        out = np.zeros((len(e), 3, 3), dtype=e.dtype)
    for i in range(len(e)):
        theta = np.linalg.norm(e[i, ...])
        if theta < 1e-16:
            out[i, 0, 0] = out[i, 1, 1] = out[i, 2, 2] = 1
            out[i, 0, 1:] = 0
            out[i, 1, 0] = out[i, 1, 2] = 0
            out[i, 2, :2] = 0
            continue
        w = e[i, ...] / theta
        s = np.sin(theta)
        c = 1 - np.cos(theta)
        w0 = s * w[0]
        w1 = s * w[1]
        w2 = s * w[2]
        w00 = -w[0] ** 2
        w11 = -w[1] ** 2
        w22 = -w[2] ** 2
        w01 = c * w[0] * w[1]
        w02 = c * w[0] * w[2]
        w12 = c * w[1] * w[2]
        out[i, 0, 0] = 1 + c * (w22 + w11)
        out[i, 0, 1] = w2 + w01
        out[i, 0, 2] = -w1 + w02
        out[i, 1, 0] = -w2 + w01
        out[i, 1, 1] = 1 + c * (w22 + w00)
        out[i, 1, 2] = w0 + w12
        out[i, 2, 0] = w1 + w02
        out[i, 2, 1] = -w0 + w12
        out[i, 2, 2] = 1 + c * (w11 + w00)
    return out


@numba.jit(nopython=True, nogil=True)
def aa2rot(e):
    return expmap(e)


@numba.jit(nopython=True, nogil=True)
def logmap(r):
    angle = np.arccos((np.trace(r) - 1) * 0.5)
    ax = np.zeros(3, dtype=r.dtype)
    if angle < 1e-12:
        return ax
    if np.abs(angle - np.pi) < 1e-12:
        ax[:3] = np.sqrt(0.5 * (np.diag(r) + 1.0))
    else:
        maginv = 0.5 / np.sin(angle)
        ax[0] = r[1, 2] - r[2, 1]
        ax[1] = r[2, 0] - r[0, 2]
        ax[2] = r[0, 1] - r[1, 0]
        ax *= maginv
        ax *= angle
    return ax


@numba.jit(nopython=True, nogil=True)
def rot2aa(r):
    return logmap(r)


def parallel_convert_func(f):
    @numba.jit(nopython=True, parallel=True)
    def g(arr, out):
        for i in numba.prange(len(arr)):
            out[i] = f(arr[i])
        return out
    return g


@numba.jit(nopython=True, nogil=True, fastmath=True)
def e2r_vec(eu, out=None):
    eu = np.atleast_2d(eu)
    if out is None:
        out = np.zeros((len(eu), 3, 3))
    for i in range(len(eu)):
        ca = np.cos(eu[i, 0])
        cb = np.cos(eu[i, 1])
        cg = np.cos(eu[i, 2])
        sa = np.sin(eu[i, 0])
        sb = np.sin(eu[i, 1])
        sg = np.sin(eu[i, 2])
        cc = cb * ca
        cs = cb * sa
        sc = sb * ca
        ss = sb * sa
        out[i, 0, :] = cg * cc - sg * sa, cg * cs + sg * ca, -cg * sb
        out[i, 1, :] = -sg * cc - cg * sa, -sg * cs + cg * ca, sg * sb
        out[i, 2, :] = sc, ss, cb
    return out


@numba.jit(nopython=True, nogil=True, fastmath=True)
def e2q_vec(eu, out=None):
    eu = np.atleast_2d(eu)
    if out is None:
        out = np.zeros((len(eu), 4))
    for i in range(len(eu)):
        out[i, 0] = np.cos((eu[i, 0] + eu[i, 2]) / 2) * np.cos(eu[i, 1] / 2)
        out[i, 1] = np.cos((eu[i, 0] - eu[i, 2]) / 2) * np.sin(eu[i, 1] / 2)
        out[i, 2] = np.sin((eu[i, 0] - eu[i, 2]) / 2) * np.sin(eu[i, 1] / 2)
        out[i, 3] = np.sin((eu[i, 0] + eu[i, 2]) / 2) * np.cos(eu[i, 1] / 2)
    return out



# from models.get_transformers import to_int8
from src.data_preprocess.mrc_preprocess import to_int8
import numpy as np
# from PIL import Image
from torch.utils.data import Dataset
# from torchvision import transforms
import os
import pickle
import random
# import mrcfile
# from PIL import Image
import torch
from io import BytesIO
import json
import lmdb
from annoy import AnnoyIndex


# import time
# from torch.utils.data import DataLoader
# from memory_profiler import profile


class MyEmFile(object):
    def __init__(self, emfile_path=None, selected_emfile_path=None, filetype='star'):
        self.filetype = filetype
        if emfile_path:
            if emfile_path.endswith(".star"):
                self.particles_file_content, self.particles_star_title, self.particles_id = self.read_star(emfile_path)
                if selected_emfile_path is not None and selected_emfile_path.endswith(".star"):
                    self.selected_particles_file_content, self.selected_particles_star_file_title, self.selected_particles_id = self.read_star(
                        selected_emfile_path)
                    self.unselected_particles_file_content, self.unselected_particles_id = self.divide_selected_unselected_particles_star(
                        self.particles_file_content, self.particles_id, self.selected_particles_id)
                else:
                    self.selected_particles_id = None

            if emfile_path.endswith(".cs"):
                self.filetype = 'cs'
                self.particles_file_content, self.particles_id = self.read_cs(emfile_path)
                if selected_emfile_path is not None and selected_emfile_path.endswith(".cs"):
                    self.selected_particles_csfile_content, self.selected_particles_id = self.read_cs(
                        selected_emfile_path)
                    self.unselected_particles_csfile_content, self.unselected_particles_id = self.divide_selected_unselected_particles_cs(
                        self.particles_file_content, self.particles_id, self.selected_particles_id)
                    # pass
                else:
                    self.selected_particles_id = None
        else:
            self.particles_id = None
            self.filetype = None
            self.selected_particles_id = None
        # pass

    def read_star(self, star_path):
        with open(star_path, "r") as starfile:
            star_data = starfile.readlines()
        for index, x in enumerate(star_data):
            if x == 'data_particles\n':
                for index2, x2 in enumerate(star_data[index:]):

                    splited_x = x2.split()
                    next_splited_x = star_data[index + index2 + 1].split()
                    if splited_x:
                        item_num = splited_x[-1].replace("#", "")
                        if item_num.isdigit():
                            if int(item_num) == len(next_splited_x) and int(item_num) != len(splited_x):
                                start_site = index + index2 + 1
                                break
        content = star_data[start_site:]
        title = star_data[:start_site]
        img_ids = self.get_star_image_id(content)
        return content, title, img_ids

    def read_cs(self, cs_path):
        cs_data = Dataset.load(cs_path)
        img_ids = cs_data['uid'].tolist()
        # mm=cs_data['blob/path'].tolist()
        # dd=cs_data['blob/idx'].tolist()
        return cs_data, img_ids

    def get_star_image_id(self, star_content):
        image_id = []
        for x in star_content:
            if len(x.strip()) > 0:
                image_id.append(x.strip().split()[5])
        return image_id

    def divide_selected_unselected_particles_star(self, particles_star_content, particles_id, selected_particles_id):
        unselected_particles_star_content = [particles_star_content[index] for index, x in enumerate(particles_id) if
                                             len(x) > 0 and x not in selected_particles_id]
        unselected_particles_id = self.get_star_image_id(unselected_particles_star_content)
        return unselected_particles_star_content, unselected_particles_id

    def divide_selected_unselected_particles_cs(self, particles_cs_content, particles_id, selected_particles_id):
        unselected_list = []
        unselected_particles_id = []
        for i, id in enumerate(particles_id):
            if id not in selected_particles_id:
                unselected_list.append(i)
                unselected_particles_id.append(id)
        unselected_particles_cs_content = particles_cs_content.take(unselected_list)
        return unselected_particles_cs_content, unselected_particles_id


class MyMrcData(MyEmFile):
    def __init__(self, cfg, mrc_path=None, emfile_path=None, processed_data_path=None, selected_emfile_path=None,
                 tmp_data_save_path=None,
                 is_extra_valset=False, accelerator=None, ctf_correction_averages=False,
                 ctf_correction_inference=False):
        super(MyMrcData, self).__init__(emfile_path, selected_emfile_path)

        self.processed_data_path = processed_data_path
        self.id_index_dict = None
        if processed_data_path is not None:
            # with open(os.path.join(processed_data_path, 'path_divided_by_labels.data'), 'rb') as filehandle:
            #     self.path_divided_by_labels = pickle.load(filehandle)
            # with open(os.path.join(processed_data_path, 'ids_divided_by_labels.data'), 'rb') as filehandle:
            #     self.ids_divided_by_labels = pickle.load(filehandle)
            self.load_preprocessed_data_path(data_path=processed_data_path,
                                             ctf_correction_averages=ctf_correction_averages,
                                             ctf_correction_train=ctf_correction_inference)
        else:
            if is_extra_valset:
                tmp_data_save_path = tmp_data_save_path + '/tmp/preprocessed_data/extra_valset/'
            else:
                tmp_data_save_path = tmp_data_save_path + '/tmp/preprocessed_data/trainset/'
            self.mrc_path = mrc_path
            if not os.path.exists(tmp_data_save_path + '/output_tif_path.data') and accelerator.is_local_main_process:
                self.load_path()
                self.divided_into_single_mrc(tmp_data_save_path, resize=cfg['resize'], crop_ratio=cfg['crop_ratio'])
            accelerator.wait_for_everyone()
            self.load_preprocessed_data_path(data_path=tmp_data_save_path)

    def load_path(self):
        mrcs_path_list = []
        listdir(self.mrc_path, mrcs_path_list)
        self.mrcs_path_list = mrcs_path_list
        # if self.selected_particles_id is not None:
        # pass

    def load_preprocessed_data_path(self, data_path, ctf_correction_averages, ctf_correction_train):
        # path_out = path_result_dir + '/tmp/preprocessed_data/'
        if os.path.exists(data_path + '/output_tif_path.data'):
            with open(data_path + '/output_tif_path.data', 'rb') as filehandle:
                self.all_tif_path = pickle.load(filehandle)
        else:
            self.all_tif_path = None
        if ctf_correction_averages and os.path.exists(data_path + '/output_ctf_tif_path.data'):
            with open(data_path + '/output_ctf_tif_path.data', 'rb') as filehandle:
                self.all_tif_path_ctf_correction = pickle.load(filehandle)
        else:
            self.all_tif_path_ctf_correction = None

        if os.path.exists(data_path + '/lmdb_data'):
            lmdb_env = lmdb.open(
                data_path + '/lmdb_data/',
                readonly=True,
                lock=False,
                readahead=False
            )
            # self.lmdb_env = lmdb_env
            processed_tif_txn = lmdb_env.begin()
            self.length = processed_tif_txn.stat()['entries']
            lmdb_env.close()
            self.lmdb_path = data_path + '/lmdb_data/'
            self.all_processed_tif_path = None
        else:
            # self.processed_tif_txn = None
            self.lmdb_env = None
            self.lmdb_path = None

            with open(data_path + '/output_processed_tif_path.data',
                      'rb') as filehandle:
                self.all_processed_tif_path = pickle.load(filehandle)
            self.length = len(self.all_processed_tif_path)
        if ctf_correction_train and os.path.exists(data_path + '/output_ctf_processed_tif_path.data'):
            with open(data_path + '/output_ctf_processed_tif_path.data', 'rb') as filehandle:
                self.all_processed_tif_path_ctf_correction = pickle.load(filehandle)
        else:
            self.all_processed_tif_path_ctf_correction = None

        if os.path.exists(data_path + '/labels_for_clustering.data'):
            with open(data_path + '/labels_for_clustering.data', 'rb') as filehandle:
                self.labels_for_clustering = pickle.load(filehandle)
        else:
            self.labels_for_clustering = None

        if os.path.exists(data_path + '/labels_classification.data'):
            with open(data_path + '/labels_classification.data', 'rb') as filehandle:
                self.labels_classification = pickle.load(filehandle)
        else:
            self.labels_classification = [-1] * self.length

        # with open(path_out + 'output_tif_select_label.data', 'rb') as filehandle:
        #     self.tifs_selection_label = pickle.load(filehandle)

        if os.path.exists(data_path + '/means_stds.data'):
            with open(data_path + '/means_stds.data',
                      'rb') as filehandle:
                self.means_stds = pickle.load(filehandle)

        if os.path.exists(data_path + '/protein_id_list.data'):
            with open(data_path + '/protein_id_list.data',
                      'rb') as filehandle:
                self.protein_id_list = pickle.load(filehandle)
        else:
            self.protein_id_list = None

        if os.path.exists(data_path + '/protein_id_dict.data'):
            with open(data_path + '/protein_id_dict.data',
                      'rb') as filehandle:
                self.protein_id_dict = pickle.load(filehandle)
        else:
            self.protein_id_dict = None

        if os.path.exists(data_path + '/pretrain_data.json'):
            self.dataset_map = json.load(open(data_path + '/pretrain_data.json', 'r'))
        elif os.path.exists(data_path + '/finetune_data.json'):
            self.dataset_map = json.load(open(data_path + '/finetune_data.json', 'r'))
        else:
            self.dataset_map = None

        if os.path.exists(data_path + '/mean_error_dict.json'):
            mean_error_dict = json.load(open(data_path + '/mean_error_dict.json', 'r'))
            mean_error_dis_dict = get_mean_error_distribution(mean_error_dict)
            self.mean_error_dis_dict = {self.protein_id_dict[key]: value for key, value in mean_error_dis_dict.items()}
            # self.mean_error_dis_dict = {'good':{self.protein_id_dict[key]: value for key, value in mean_error_dis_dict['good'].items()},
            #                             'bad':{self.protein_id_dict[key]: value for key, value in mean_error_dis_dict['bad'].items()}}
        else:
            self.mean_error_dis_dict = None

        if os.path.exists(data_path + '/data_error_dict.json'):
            data_error_dict = json.load(open(data_path + '/data_error_dict.json', 'r'))
            self.data_error_dis_dict = {self.protein_id_dict[key]: np.array(value) for key, value in
                                        data_error_dict.items()}

        if os.path.exists(data_path + '/pose_id_map.data'):
            # self.pose_id_map = json.load(open(data_path + '/pose_id_map.data', 'r'))
            with open(data_path + '/pose_id_map.data',
                      'rb') as filehandle:
                self.pose_id_map = pickle.load(filehandle)
        else:
            self.pose_id_map = None

        # with open(data_path + '/labels_for_training.data',
        #           'rb') as filehandle:
        #     self.labels_for_training = pickle.load(filehandle)
        # with open(data_path + '/probabilities_for_sampling.data',
        #           'rb') as filehandle:
        #     self.probabilities_for_sampling = pickle.load(filehandle)

    # def preprocess_trainset_valset_index(self, valset_name=[], dataset_except_names=[], is_balance=False,
    #                                      max_resample_num=None, max_resample_num_val=None, ratio_balance_train=[0.35,0.3,0.35],
    #                                      max_number_per_sample=None):
    #     valset_name_id = [self.protein_id_dict[name] for name in valset_name]
    #     dataset_except_names_id = [self.protein_id_dict[name] for name in dataset_except_names]
    #     id_sum_dict = {id: 0 for id in self.protein_id_dict.values()}
    #     dataset_index = []
    #     valset_index = []
    #     positive_index = []
    #     middle_index = []
    #     negative_index = []
    #     resample_num_p = 0
    #     resample_num_m=0
    #     resample_num_n = 0
    #     if len(valset_name_id) > 0 or len(dataset_except_names_id) > 0:
    #         for i, protein_id in enumerate(self.protein_id_list):
    #             if protein_id in valset_name_id:
    #                 valset_index.append(i)
    #             elif protein_id not in dataset_except_names_id:
    #                 dataset_index.append(i)
    #     else:
    #         # dataset_index = list(range(len(self.all_processed_tif_path)))
    #         dataset_index = list(range(self.length))
    #     if is_balance:
    #         for i in dataset_index:
    #             if self.labels_classification[i] == 1:
    #                 if max_number_per_sample is not None:
    #                     if id_sum_dict[self.protein_id_list[i]] < max_number_per_sample:
    #                         positive_index.append(i)
    #                         id_sum_dict[self.protein_id_list[i]] += 1
    #                 else:
    #                     positive_index.append(i)
    #             else:
    #                 if max_number_per_sample is not None:
    #                     if id_sum_dict[self.protein_id_list[i]] < max_number_per_sample:
    #                         negative_index.append(i)
    #                         id_sum_dict[self.protein_id_list[i]] += 1
    #                 else:
    #                     negative_index.append(i)
    #         if len(positive_index) > len(negative_index):
    #             resample_num = len(negative_index)
    #             # positive_index=random.sample(positive_index,len(negative_index))
    #         else:
    #             resample_num = len(positive_index)
    #             # negative_index=random.sample(negative_index,len(positive_index))
    #         if max_resample_num is not None:
    #             resample_num_p = int(max_resample_num * ratio_balance_train[1]) if max_resample_num * ratio_balance_train[1] < len(
    #                 positive_index) else resample_num
    #             resample_num_n = max_resample_num - resample_num_p
    #         else:
    #             resample_num_p = resample_num
    #             resample_num_n = resample_num
    #
    #         sub_positive_index = random.sample(positive_index, resample_num_p)
    #         sub_negative_index = random.sample(negative_index, resample_num_n)
    #         # if len(positive_index)>len(negative_index):
    #         #     positive_index=random.sample(positive_index,len(negative_index))
    #         # else:
    #         #     negative_index=random.sample(negative_index,len(positive_index))
    #         dataset_index = sub_positive_index + sub_negative_index
    #         if max_resample_num_val is not None:
    #             if len(valset_index) > max_resample_num_val:
    #                 valset_index = random.sample(valset_index, max_resample_num_val)
    #     elif max_resample_num is not None:
    #         if len(dataset_index) > max_resample_num:
    #             dataset_index = random.sample(dataset_index, max_resample_num)
    #     return dataset_index, valset_index, positive_index, negative_index, (resample_num_p, resample_num_n)

    # def preprocess_trainset_valset_index_finetune(self, valset_name=[], dataset_except_names=[],
    #                                               positive_ratio=0.5,
    #                                               max_number_per_sample=None, is_valset=False):
    #     if is_valset:
    #         id_index_dict_pos = {id: [] for name, id in self.protein_id_dict.items() if name.lower().endswith(
    #             'good') and name in valset_name and name not in dataset_except_names}
    #         id_index_dict_neg = {id: [] for name, id in self.protein_id_dict.items() if name.lower().endswith(
    #             'bad') and name in valset_name and name not in dataset_except_names}
    #     else:
    #         id_index_dict_pos = {id: [] for name, id in self.protein_id_dict.items() if name.lower().endswith(
    #             'good') and name not in valset_name and name not in dataset_except_names}
    #         id_index_dict_neg = {id: [] for name, id in self.protein_id_dict.items() if name.lower().endswith(
    #             'bad') and name not in valset_name and name not in dataset_except_names}
    #     protein_id_list_np = np.array(self.protein_id_list)
    #     for name, id in self.protein_id_dict.items():
    #         if name.lower().endswith('good'):
    #             id_index_dict_pos[id] = np.where(protein_id_list_np == id)[0].tolist()
    #         elif name.lower().endswith('bad'):
    #             id_index_dict_neg[id] = np.where(protein_id_list_np == id)[0].tolist()
    #     resample_num_p = int(max_number_per_sample * 4 * positive_ratio * len(id_index_dict_neg) / (
    #                 len(id_index_dict_pos) + len(id_index_dict_neg)))
    #     resample_num_n = int(max_number_per_sample * 2 - resample_num_p)
    #     return id_index_dict_pos, id_index_dict_neg, (resample_num_p, resample_num_n)

    def preprocess_trainset_valset_index_finetune(self,
                                                  # valset_name=[],
                                                  # dataset_except_names=[],
                                                  ratio_balance_train=[0.35, 0.3, 0.35],
                                                  max_number_per_sample=None,
                                                  # is_valset=False,
                                                  is_balance=True,
                                                  middle_range_balance_train=[0.5, 0.85],
                                                  # data_error_dis_dict=None
                                                  ):
        id_index_dict_pos = {}
        id_index_dict_neg = {}
        id_index_dict_mid = {}

        # if data_error_dis_dict is not None:
        #     data_error_dis_dict_pos = {}
        #     data_error_dis_dict_neg = {}
        #     data_error_dis_dict_mid = {}

        protein_id_list_np = np.array(self.protein_id_list)
        labels_classification_np = np.array(self.labels_classification)
        for name, id in self.protein_id_dict.items():
            item_pos={}
            item_neg={}
            item_mid={}
            # if name not in dataset_except_names:
            if name.lower().endswith('good'):
                # if name in valset_name and is_valset:
                #     id_index_dict_pos[id] = np.where(protein_id_list_np == id)[0].tolist()
                # elif name not in valset_name and not is_valset:
                #     id_index_dict_pos[id] = np.where(protein_id_list_np == id)[0].tolist()
                #     if data_error_dis_dict is not None:
                #         data_error_dis_dict_pos[id] = data_error_dis_dict[id] / np.sum(data_error_dis_dict[id])
                item_pos['id'] = np.where(protein_id_list_np == id)[0].tolist()
                item_pos['score']=[1.0]*len(item_pos['id'])
            elif name.lower().endswith('bad'):
                # if name in valset_name and is_valset:
                #     id_index_dict_neg[id] = np.where(protein_id_list_np == id)[0].tolist()
                # elif name not in valset_name and not is_valset:
                #     id_index_dict_neg[id] = np.where(protein_id_list_np == id)[0].tolist()
                #     if data_error_dis_dict is not None:
                #         data_error_dis_dict_neg[id] = data_error_dis_dict[id] / np.sum(data_error_dis_dict[id])
                item_neg['id'] = np.where(protein_id_list_np == id)[0].tolist()
                item_neg['score']=[0.0]*len(item_neg['id'])
            else:
                protein_index = np.where(protein_id_list_np == id)[0]
                pos_index = protein_index[labels_classification_np[protein_index] >= middle_range_balance_train[1]]
                neg_index = protein_index[labels_classification_np[protein_index] < middle_range_balance_train[0]]
                # if data_error_dis_dict is not None:
                #     pos_dis = data_error_dis_dict[id][
                #         labels_classification_np[protein_index] >= middle_range_balance_train[1]]
                #     neg_dis = data_error_dis_dict[id][
                #         labels_classification_np[protein_index] < middle_range_balance_train[0]]
                if middle_range_balance_train[0] != middle_range_balance_train[1]:
                    mid_index = protein_index[
                        (labels_classification_np[protein_index] >= middle_range_balance_train[0]) & (
                                labels_classification_np[protein_index] < middle_range_balance_train[1])]
                    if len(mid_index) > 0:
                        # id_index_dict_mid[id] = mid_index.tolist()
                        item_mid['id'] = mid_index.tolist()
                        item_mid['score'] = labels_classification_np[mid_index].tolist()
                        # if data_error_dis_dict is not None:
                        #     mid_dis = data_error_dis_dict[id][
                        #         (labels_classification_np[protein_index] >= middle_range_balance_train[0]) & (
                        #                 labels_classification_np[protein_index] < middle_range_balance_train[1])]
                        #     data_error_dis_dict_mid[id] = mid_dis / np.sum(mid_dis)
                    # id_index_dict_mid[id] = mid_index.tolist()
                # if name in valset_name and is_valset:
                #     id_index_dict_pos[id] = pos_index.tolist()
                #     id_index_dict_neg[id] = neg_index.tolist()
                #
                # elif name not in valset_name and not is_valset:
                #     if len(pos_index) > 0:
                #         id_index_dict_pos[id] = pos_index.tolist()
                #         if data_error_dis_dict is not None:
                #             data_error_dis_dict_pos[id] = pos_dis / np.sum(pos_dis)
                #     if len(neg_index) > 0:
                #         id_index_dict_neg[id] = neg_index.tolist()
                #         if data_error_dis_dict is not None:
                #             data_error_dis_dict_neg[id] = neg_dis / np.sum(neg_dis)
                if len(pos_index) > 0:
                    item_pos['id'] = pos_index.tolist()
                    item_pos['score'] = labels_classification_np[pos_index].tolist()
                if len(neg_index) > 0:
                    item_neg['id'] = neg_index.tolist()
                    item_neg['score'] = labels_classification_np[neg_index].tolist()
            if len(item_pos) > 0:
                id_index_dict_pos[id] = item_pos
            if len(item_neg) > 0:
                id_index_dict_neg[id] = item_neg
            if len(item_mid) > 0:
                id_index_dict_mid[id] = item_mid
        if is_balance:
            resample_num_p = int(max_number_per_sample * ratio_balance_train[2])
            resample_num_n = int(max_number_per_sample * ratio_balance_train[0])
            resample_num_m = int(max_number_per_sample * ratio_balance_train[1]) if len(id_index_dict_mid) > 0 else 0
            # if len(id_index_dict_mid)==0:
            #     resample_num_p = int(max_number_per_sample * 4 * ratio_balance_train[1] * len(id_index_dict_neg) / (
            #             len(id_index_dict_pos) + len(id_index_dict_neg)))
            #     resample_num_n = int(max_number_per_sample * 2 - resample_num_p)
            #     resample_num_m=0
            # else:
            #     # resample_num_p=int(6 * ratio_balance_train[1] * max_number_per_sample * (len(id_index_dict_neg) * len(id_index_dict_mid)) / (len(id_index_dict_pos) * len(id_index_dict_neg) + len(id_index_dict_neg) * len(id_index_dict_mid) + len(id_index_dict_pos) * len(id_index_dict_mid)))
            #     # resample_num_n=int(3*(1-positive_ratio)*max_number_per_sample*(len(id_index_dict_pos)*len(id_index_dict_mid))/(len(id_index_dict_pos)*len(id_index_dict_neg)+len(id_index_dict_neg)*len(id_index_dict_mid)+len(id_index_dict_pos)*len(id_index_dict_mid)))
            #     # resample_num_m=2*max_number_per_sample-resample_num_p-resample_num_n
            #     # resample_num_p = int(2 * positive_ratio * max_number_per_sample)
            #     # resample_num_n = int( max_number_per_sample-resample_num_p/2)
            #     # resample_num_m = 2*max_number_per_sample-resample_num_p-resample_num_n
            #     # ratio_multi=[1,1,1]
            #     resample_num_p = int(max_number_per_sample* ratio_balance_train[2])
            #     resample_num_n = int(max_number_per_sample* ratio_balance_train[0])
            #     resample_num_m = int(max_number_per_sample* ratio_balance_train[1])

        else:
            resample_num_p = max_number_per_sample
            resample_num_n = max_number_per_sample
            resample_num_m = max_number_per_sample
        # data_error_dis_dict_all = {'good': data_error_dis_dict_pos, 'bad': data_error_dis_dict_neg,
        #                            'mid': data_error_dis_dict_mid} if data_error_dis_dict is not None else {
        #     'good': None, 'bad': None, 'mid': None}
        return id_index_dict_pos, id_index_dict_neg, id_index_dict_mid, (
            resample_num_p, resample_num_n, resample_num_m)

    def preprocess_trainset_index_pretrain(self, protein_id_dict=None, protein_id_list=None):
        if protein_id_dict is not None and protein_id_list is not None:
            target_protein_id_dict = protein_id_dict
            target_protein_id_list = protein_id_list
        else:
            target_protein_id_dict = self.protein_id_dict
            target_protein_id_list = self.protein_id_list
        bad_id_list_all = [target_protein_id_dict[name] for name in target_protein_id_dict.keys() if
                           name.lower().endswith('bad')]
        if self.dataset_map is None:

            # dataset_id_map=None
            id_map = None
            bad_id_list = None
        else:
            # id_index_dict = {target_protein_id_dict[name]: [] for name in self.dataset_map.keys()}
            id_map = {target_protein_id_dict[name]: target_protein_id_dict[name2] if name2 is not None else None for
                      name, name2 in self.dataset_map.items()}
            bad_id_list = [target_protein_id_dict[name] for name in self.dataset_map.keys() if
                           name.lower().endswith('bad')]

        dataset_id_map = {'id_map': id_map, 'bad_id_list': bad_id_list, 'bad_id_list_all': bad_id_list_all}
        # for i, id in enumerate(self.protein_id_list):
        #     id_index_dict[id].append(i)
        id_index_dict = {id: [] for id in target_protein_id_dict.values()}
        id_scores_dict= {}
        scores_np= np.array(self.labels_classification)
        protein_id_list_np = np.array(target_protein_id_list)
        for id in target_protein_id_dict.values():
            # aaa = np.where(protein_id_list_np == id)
            id_index_dict[id] = np.where(protein_id_list_np == id)[0].tolist()
            id_scores_dict[id] = scores_np[id_index_dict[id]]
        self.id_index_dict = id_index_dict
        return id_index_dict, dataset_id_map,id_scores_dict


class EMDataset_from_path(Dataset):
    """自定义数据集"""

    def __init__(self, mrcdata: MyMrcData, transform=None,
                 normal_scale=10, accelerator=None,
                 local_crops=None,
                 slice_setting=None,
                 mix_pos_setting=None,
                 weight_for_contrastive_classification_label=0.0,
                 use_triplex_labels=False, bar_score=0.4,
                 in_chans=1, needs_aug2=True
                 ):
        self.pose_indices = AnnoyIndex(2, 'euclidean')
        self.tif_len = mrcdata.length
        self.lmdb_path = mrcdata.lmdb_path
        # if mrcdata.lmdb_path is not None:
        #     self.lmdb_env=lmdb.open(
        #         mrcdata.lmdb_path,
        #         readonly=True,
        #         # lock=False,
        #         # readahead=False
        #     )
        #     self.processed_tif_txn = self.lmdb_env.begin()
        # # if mrcdata.lmdb_env is not None:
        # #     self.processed_tif_txn = mrcdata.lmdb_env.begin()
        # else:
        #     self.processed_tif_txn = None
        # self.processed_tif_txn = mrcdata.processed_tif_txn
        self.tif_path_list = mrcdata.all_processed_tif_path
        self.tif_path_list_ctf_correction = mrcdata.all_processed_tif_path_ctf_correction
        self.tif_path_list_raw = mrcdata.all_tif_path
        self.tif_path_list_raw_ctf_correction = mrcdata.all_tif_path_ctf_correction
        self.labels_for_clustering = mrcdata.labels_for_clustering
        self.labels_classification = mrcdata.labels_classification
        self.id_index_dict = mrcdata.id_index_dict

        self.slice_setting = slice_setting
        self.mix_pos_setting = mix_pos_setting
        self.in_chans = in_chans
        self.needs_aug2 = needs_aug2

        self.use_triplex_labels = use_triplex_labels
        if bar_score < 0.5:
            self.bar_score = 1 - bar_score
        else:
            self.bar_score = bar_score

        self.protein_id_list = mrcdata.protein_id_list
        self.protein_id_dict = mrcdata.protein_id_dict
        self.protein_id_dict_reverse = {v: k for k, v in self.protein_id_dict.items()}

        if mrcdata.particles_id is not None:
            self.particles_id = mrcdata.particles_id
        else:
            self.particles_id = range(self.tif_len)
        # self.isnorm = is_Normalize
        # self.mean_std = mrcdata.means_stds
        self.normal_scale = normal_scale
        self.transform = transform
        self.accelerator = accelerator
        # self.train=True

        # self.labels_for_training = mrcdata.labels_for_training
        # self.probabilities_for_sampling = mrcdata.probabilities_for_sampling
        self.processed_data_path = mrcdata.processed_data_path

        if local_crops is not None:
            self.local_crops_number = local_crops['number']
        else:
            self.local_crops_number = 0

        self.weight_for_contrastive_classification_label = weight_for_contrastive_classification_label
        if weight_for_contrastive_classification_label > 0:
            labels_classification_np = np.array(self.labels_classification)
            self.positive_items = np.where(labels_classification_np == 1)[0]
            self.negative_items = np.where(labels_classification_np == 0)[0]
        if slice_setting is not None and (slice_setting['p'] > 0 or slice_setting['align_p'] > 0) and slice_setting[
            'processed_path_slice'] is not None:
            with open(slice_setting['processed_path_slice'] + '/output_processed_tif_path.data', 'rb') as filehandle:
                self.tif_path_list_slice = pickle.load(filehandle)
        else:
            self.tif_path_list_slice = None

        self.pose_id_map = mrcdata.pose_id_map

    def __len__(self):
        return self.tif_len
        # return len(self.tif_path_list)

    def open_lmdb(self):
        # if mrcdata.lmdb_path is not None:
        self.lmdb_env = lmdb.open(
            self.lmdb_path,
            readonly=True,
            meminit=False,
            max_readers=1,
            lock=False,
            readahead=False
        )

    # @profile(precision=4)
    def __getitem__(self, item):
        local_crops1 = []
        local_crops2 = []
        # end_git_item=time.time()

        '''get mrcdata1 and aug1'''
        mrcdata = self.get_mrcdata(item=item)

        weight = float(1.0)
        # label_for_training=self.labels_for_training[item]
        # gaussian_probabilities=self.probabilities_for_sampling[item]

        '''get labels data'''
        if self.labels_for_clustering is not None and len(self.labels_for_clustering) > item:
            label_for_clustering = self.labels_for_clustering[item]
        else:
            label_for_clustering = -1
        label_for_classification = self.labels_classification[item]
        if self.use_triplex_labels:
            if label_for_classification > self.bar_score:
                label_for_classification = 1.0
            elif label_for_classification <= 1 - self.bar_score:
                label_for_classification = 0.0
            else:
                label_for_classification = 0.5

        # particles_id = self.particles_id[item]
        protein_id = self.protein_id_list[item]

        '''get mrcdata2'''
        if self.needs_aug2:
            mrcdata2 = None
            is_random_rotate_transform = True if self.random_rotate_transform is not None else False
            item2, weight,is_mix_pos = self.get_item2(item)
            if item2 is not None:
                mrcdata2 = self.get_mrcdata(item=item2)
            elif self.slice_setting is not None and random.random() < self.slice_setting['p']:
                mrcdata2 = self.get_corr_slice(item)
            if mrcdata2 is None:
                mrcdata2 = mrcdata
            # else:
            #     is_random_rotate_transform = False
            aug1 = self.mrcdata_aug(mrcdata, is_random_rotate_transform=is_random_rotate_transform,is_mix_pos=is_mix_pos)
            aug2 = self.mrcdata_aug(mrcdata2, is_random_rotate_transform=is_random_rotate_transform,is_mix_pos=is_mix_pos)
        else:
            aug1 = self.mrcdata_aug(mrcdata)
            aug2 = None

        # img2tensor = transforms.ToTensor()
        # mrcdata = img2tensor(mrcdata)

        out = {
            # 'mrcdata': mrcdata,
            'aug1': aug1, 'aug2': aug2 if aug2 is not None else [],
            'weight': weight,
            'label_for_clustering': label_for_clustering,
            'label_for_classification': label_for_classification,
            # 'path': tif_path,
            # 'raw_path': raw_tif_path,
            # 'particles_id': str(particles_id),
            'item': item,
            'local_crops1': local_crops1,
            'local_crops2': local_crops2, 'protein_id': protein_id}
        return out

    def get_item2(self, item):
        item2 = None
        weight = 1
        is_mix_pos=False
        if self.weight_for_contrastive_classification_label > 0:
            if random.random() < self.weight_for_contrastive_classification_label:
                if self.labels_classification[item] == 1:
                    item2 = np.random.choice(self.positive_items)
                else:
                    item2 = np.random.choice(self.negative_items)
        elif self.mix_pos_setting is not None and self.mix_pos_setting['p'] > 0 and self.pose_id_map is not None:
            if random.random() < self.mix_pos_setting['p']:
                # protein_id = self.protein_id_list[item]
                # item_list=self.id_index_dict.get(protein_id, [])
                # if len(item_list) > 1:
                #     item2 = random.choice(item_list)
                # protein_name=self.protein_id_dict_reverse[protein_id]
                # self.pose_indices.load(os.path.join(self.processed_data_path, 'pose_data', protein_name + '_pose.ann'))
                # nearest=self.pose_indices.get_nns_by_item(item-min(item_list), int(len(item_list)/20),include_distances=False)
                # item2 = random.choice(nearest[1:])+ min(item_list)
                nearest, min_id, protein_name, pose_items_id,item1_pose_id = self.get_nearest_item(item)
                if nearest is not None and len(nearest) > 1:
                    item2_pose_id=weighted_random_choice_linear(nearest[1:],with_weight=False)
                    item2 = pose_items_id[item2_pose_id]+ min_id

                    # if item-min(item_list)<0:
                    #     print('item is less than min(item_list): ' + str(item) + ' ' + str(min(item_list)))
                    #     print('protein_name: ' + protein_name)
                    # if item2-min(item_list)<0:
                    #     print('item2 is less than min(item_list): ' + str(item2) + ' ' + str(min(item_list)))
                    #     print('protein_name: ' + protein_name)
                    weight = sigmoid(self.mix_pos_setting['sigma'] * (
                        (3.5 - self.pose_indices.get_distance(item1_pose_id, item2_pose_id) )/3.5
                                - self.mix_pos_setting['bias']))
                    self.pose_indices.unload()
                    is_mix_pos=True

                # if protein_name=='11307_J504_good':
                #     pass
        return item2, float(weight),is_mix_pos

    def  get_nearest_item(self, item, N=None):

        protein_id = self.protein_id_list[item]
        item_list = self.id_index_dict.get(protein_id, [])
        min_id = min(item_list)
        nearest = None
        protein_name = self.protein_id_dict_reverse[protein_id]
        pose_items_id = []
        item1_pose_id = None
        if item - min_id in self.pose_id_map[protein_id]:
            item1_pose_id = self.pose_id_map[protein_id][item - min_id]
            # if len(item_list) > 1:
            #     item2 = random.choice(item_list)

            self.pose_indices.load(os.path.join(self.processed_data_path, 'pose_data', protein_name + '_pose.ann'))
            nearest = self.pose_indices.get_nns_by_item(item1_pose_id,
                                                        int(len(item_list) / 50) if N is None else N,
                                                        include_distances=False)
            pose_items_id = list(self.pose_id_map[protein_id].keys())
            # nearest=nearest[1:]
        return nearest, min_id, protein_name, pose_items_id,item1_pose_id

    def get_corr_slice(self, item):
        tif_path = self.tif_path_list[item]
        tif_path_split = tif_path.split('/')
        tif_path_split[-5] += '_slice'
        tif_path2 = '/'.join(tif_path_split)
        mrcdata2 = None
        if os.path.exists(tif_path2):
            try:
                with open(tif_path2,
                          'rb') as filehandle:
                    mrcdata2 = pickle.load(filehandle)

            except EOFError:
                print('error for path: ' + tif_path)
        return mrcdata2

    def get_mrcdata(self, item=None, tif_path=None):
        mrcdata = None
        if tif_path is not None:
            if os.path.exists(tif_path):
                try:
                    with open(tif_path,
                              'rb') as filehandle:
                        mrcdata = pickle.load(filehandle)

                except EOFError:
                    print('error for path: ' + tif_path)
        elif item is not None:
            if self.lmdb_path is not None:
                if not hasattr(self, 'lmdb_env'):
                    self.open_lmdb()
                with self.lmdb_env.begin(write=False) as txn:
                    key = f"{item}".encode()
                    value = txn.get(key)
                    data = torch.load(BytesIO(value))
                    mrcdata = data['image_processed']
                    tif_path = ''
                    # raw_tif_path = ''
                del data, value
            else:
                tif_path = self.tif_path_list[item]
                if self.slice_setting is not None and (random.random() < self.slice_setting['p']):
                    slice_path = random.choice(self.tif_path_list_slice)
                    if os.path.exists(slice_path):
                        tif_path = slice_path

                try:
                    with open(tif_path,
                              'rb') as filehandle:
                        mrcdata = pickle.load(filehandle)

                except EOFError:
                    print('error for path: ' + tif_path)
        return mrcdata

    def mrcdata_aug(self, mrcdata, is_random_rotate_transform=True,is_mix_pos=False):
        if mrcdata.mode != 'L':
            mrcdata = to_int8(mrcdata)
        if  is_random_rotate_transform:
            if is_mix_pos and self.random_rotate_transform_mix_pos is not None:
                mrcdata_rotate1 = self.random_rotate_transform_mix_pos(mrcdata)
            elif self.random_rotate_transform is not None:
                mrcdata_rotate1 = self.random_rotate_transform(mrcdata)
            else:
                mrcdata_rotate1 = mrcdata
        else:
            mrcdata_rotate1 = mrcdata
        if is_mix_pos:
            aug=self.transform_mix_pos(mrcdata_rotate1)
        else:
            aug = self.transform(mrcdata_rotate1)
        return aug

    def get_transforms(self, transforms,transforms_list_mix_pos =None):
        self.transform = transforms[0]
        self.local_crops_transform = transforms[1]
        self.random_rotate_transform = transforms[2]
        if transforms_list_mix_pos is not None:
            # self.mix_pos_transforms = transforms_list_mix_pos
            self.random_rotate_transform_mix_pos = transforms_list_mix_pos[2]
            self.transform_mix_pos = transforms_list_mix_pos[0]
        else:
            self.mix_pos_transforms = None
            self.random_rotate_transform_mix_pos = None
            self.transform_mix_pos = None

    def get_local_crops(self, aug1, aug2, mrcdata_rotate1, mrcdata_rotate2):
        local_crops1 = []
        local_crops2 = []
        for _ in range(self.local_crops_number):
            local_crops1.append(self.local_crops_transform(mrcdata_rotate1))
            if self.needs_aug2:
                local_crops2.append(self.local_crops_transform(mrcdata_rotate2))
        # imgs_all = [aug1, aug2] + local_crops1 + local_crops2

        if self.in_chans != 1:
            aug1 = aug1.repeat(self.in_chans, 1, 1)
            local_crops1 = [local_crop.repeat(self.in_chans, 1, 1) for local_crop in local_crops1]
            if self.needs_aug2:
                aug2 = aug2.repeat(self.in_chans, 1, 1)
                local_crops2 = [local_crop.repeat(self.in_chans, 1, 1) for local_crop in local_crops2]
        return local_crops1, local_crops2


def listdir(path, list_name):  # 传入存储的list
    for file in os.listdir(path):
        file_path = os.path.join(path, file)
        if os.path.isdir(file_path):
            listdir(file_path, list_name)
        elif os.path.splitext(file)[-1] == '.mrc' or os.path.splitext(file)[-1] == '.mrcs':
            list_name.append(file_path)


def sigmoid(x):
    return 1.0 / (1 + np.exp(-x))


def weighted_random_choice_linear(my_list,with_weight=True):
    """
    使用线性递减的权重从列表中随机抽取一个元素。
    """

    if not my_list:
        return None

    if not with_weight:
        # 如果不需要权重，直接随机选择
        return random.choice(my_list)

    # 1. 生成权重列表 [len(my_list), len(my_list)-1, ..., 1]
    list_length = len(my_list)
    weights = list(range(list_length, 0, -1))
    # 或者使用列表推导式: weights = [list_length - i for i in range(list_length)]

    # 2. 使用 random.choices 进行加权抽样
    # k=1 表示只抽取一个元素，返回的是一个列表，所以用 [0] 获取该元素
    return random.choices(my_list, weights=weights, k=1)[0]



import pandas as pd
import os
from .generate_cs_file import generate_cs_file
from .cs2star import cs2star
from cryosparc.dataset import Dataset

def listdir(path, list_name):  # 传入存储的list
    for file in os.listdir(path):
        file_path = os.path.join(path, file)
        if os.path.isdir(file_path):
            listdir(file_path, list_name)
        elif os.path.splitext(file)[-1] == '.csv':
            list_name.append(file_path)


def read_record_data(path,target_mrcs_path= None):
    path_list = []
    blob_id_list = []
    record_data_path_list = []
    defocus_1_list = []
    defocus_2_list = []
    pixelsize_list = []
    size_list=[]
    listdir(path, record_data_path_list)
    for record_data_path in record_data_path_list:
        record_data = pd.read_csv(record_data_path)
        path_list.extend(record_data['path'].tolist())
        blob_id_list.extend(record_data.index.tolist())
        defocus_1_list.extend(record_data['defocus_1'].tolist())
        defocus_2_list.extend(record_data['defocus_2'].tolist())
        pixelsize_list.extend(record_data['pixelsize'].tolist())
        size_list.extend(record_data['image_size'].tolist())
    size_list=[[x,x] for x in size_list]
    if target_mrcs_path is not None:
        path_list=[os.path.join(target_mrcs_path,x.split('/')[-1]) for x in path_list]
    return {'path': path_list, 'blob_id': blob_id_list, 'defocus_1': defocus_1_list, 'defocus_2': defocus_2_list,
            'pixelsize': pixelsize_list,'image_size':size_list}


def write_record_data(path, record_data):
    cs_data = generate_cs_file(blob_path=record_data['path'], blob_id=record_data['blob_id'],
                               psize=record_data['pixelsize'], df1=record_data['defocus_1'],
                               df2=record_data['defocus_2'], image_size=record_data['image_size'])
    cs_data.save(path+'/simluated_particles.cs')
    cs2star(path+'/simluated_particles.cs', path+'/simluated_particles.star')




# Copyright (C) 2017 Daniel Asarnow
# University of California, San Francisco
#
# Handles metadata from cryoSPARC v0.65 and earlier.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
import pandas as pd
from .. import geom
from .. import star


def parse_cryosparc_065_csv(csvfile):
    with open(csvfile, 'rU') as f:
        lines = enumerate(f)
        idx = -1
        headers = None
        for i, l in lines:
            if l.startswith("_header"):
                _, headers = next(lines)
                headers = headers.rstrip().split(",")
            if l.startswith("_dtypes"):
                _i, dtypes = next(lines)
                dtypes = dtypes.rstrip().split(",")
                idx = _i + 1
                break
    meta = pd.read_csv(csvfile, skiprows=idx, header=None, skip_blank_lines=True)
    if headers is None:
        return None
    meta.columns = headers
    return meta


def cryosparc_065_csv2star(meta, minphic=0):
    general = {u'uid': None,
               u'split': "rlnRandomSubset",
               u'ctf_params.akv': "rlnVoltage",
               u'ctf_params.angast_deg': "rlnDefocusAngle",
               u'ctf_params.angast_rad': None,
               u'ctf_params.cs': "rlnSphericalAberration",
               u'ctf_params.detector_psize': "rlnDetectorPixelSize",
               u'ctf_params.df1': "rlnDefocusU",
               u'ctf_params.df2': "rlnDefocusV",
               u'ctf_params.mag': "rlnMagnification",
               u'ctf_params.phase_shift': "rlnPhaseShift",
               u'ctf_params.psize': None,
               u'ctf_params.wgh': "rlnAmplitudeContrast",
               u'data_input_relpath': "rlnImageName",
               u'data_input_idx': None}
    model = {u'alignments.model.U': None,
             u'alignments.model.dr': None,
             u'alignments.model.dt': None,
             u'alignments.model.ess_R': None,
             u'alignments.model.ess_S': None,
             u'alignments.model.phiC': None,
             u'alignments.model.r.0': "rlnAngleRot",
             u'alignments.model.r.1': "rlnAngleTilt",
             u'alignments.model.r.2': "rlnAnglePsi",
             u'alignments.model.t.0': "rlnOriginX",
             u'alignments.model.t.1': "rlnOriginY"}
    meta["data_input_idx"] = ["%.6d" % (i + 1) for i in meta[
        "data_input_idx"]]  # Reformat particle idx for Relion.
    if "data_input_relpath" not in meta.columns:
        meta["data_input_relpath"] = ""
    meta["data_input_relpath"] = meta["data_input_idx"].str.cat(
        meta["data_input_relpath"], sep="@")  # Construct rlnImageName field.
    # Take care of trivial mappings.
    rlnheaders = [general[h] for h in meta.columns if
                  h in general and general[h] is not None]
    df = meta[[h for h in meta.columns if
               h in general and general[h] is not None]].copy()
    df.columns = rlnheaders
    df = star.augment_star_ucsf(df, inplace=True)
    if "rlnRandomSubset" in df.columns:
        df["rlnRandomSubset"] = df["rlnRandomSubset"].apply(
            lambda x: ord(x) - 64)
    if "rlnPhaseShift" in df.columns:
        df["rlnPhaseShift"] = np.rad2deg(df["rlnPhaseShift"])
    # Class assignments and other model parameters.
    phic = meta[[h for h in meta.columns if
                 "phiC" in h]]  # Posterior probability over class assignments.
    if len(phic.columns) > 0:  # Check class assignments exist in input.
        # phic.columns = [int(h[21]) for h in meta.columns if "phiC" in h]
        phic.columns = range(len(phic.columns))
        cls = phic.idxmax(axis=1)
        for p in model:
            if model[p] is not None:
                pspec = p.split("model")[1]
                param = meta[[h for h in meta.columns if pspec in h]]
                if len(param.columns) > 0:
                    param.columns = phic.columns
                    df[model[p]] = param.lookup(param.index, cls)
        df["rlnClassNumber"] = cls + 1  # Add one for Relion indexing.
    else:
        for p in model:
            if model[p] is not None and p in meta.columns:
                df[model[p]] = meta[p]
        df["rlnClassNumber"] = 1
    if df.columns.intersection(star.Relion.ANGLES).size == len(star.Relion.ANGLES):
        df[star.Relion.ANGLES] = np.rad2deg(geom.rot2euler(geom.expmap(df[star.Relion.ANGLES].values)))
    if phic is not None and minphic > 0:
        mask = np.all(phic < minphic, axis=1)
        df.drop(df[mask].index, inplace=True)
    return df



# Copyright (C) 2019 Daniel Asarnow
# University of California, San Francisco
#
# Library of miscellaneous utility functions.
# See help text and README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numba
import numpy as np


@numba.guvectorize(["void(float64[:], float64[:], float64[:])"],
                   "(m),(m)->(m)", nopython=True, cache=False)
def cross3(u, v, w):
    w[0] = u[1] * v[2] - u[2] * v[1]
    w[1] = u[2] * v[0] - u[0] * v[2]
    w[2] = u[0] * v[1] - u[1] * v[0]


@numba.jit(nopython=True, cache=False, nogil=True)
def cross3_sca(u, v):
    w = np.zeros_like(u)
    w[0] = u[1] * v[2] - u[2] * v[1]
    w[1] = u[2] * v[0] - u[0] * v[2]
    w[2] = u[0] * v[1] - u[1] * v[0]
    return w


@numba.jit(nopython=True, cache=False, nogil=True)
def dualcos(r, d):
    return np.cos(r), -d * np.sin(r)



# Copyright (C) 2018 Daniel Asarnow
# University of California, San Francisco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from .geom_numba import cross3


def qrotate(q, v):
    return v + cross3(2 * q[:, 1:], cross3(q[:, 1:], v) + q[:, 0].reshape(-1, 1) * v)


# def qslerp(q1, q2, t):
#     cos_half_theta = np.dot(q1, q2)
#     if cos_half_theta >= 1.0:
#         return q1.copy()
#     half_theta = np.arccos(cos_half_theta)
#     sin_half_theta = np.sqrt(1 - cos_half_theta * cos_half_theta)
#     if np.abs(sin_half_theta) < 1E-12:
#         return (q1 + q2) / 2
#     a = np.sin((1 - t) * half_theta) / sin_half_theta
#     b = np.sin(t * half_theta) / sin_half_theta
#     return q1 * a + q2 * b


def normq(q, mu=None):
    q = (q.T/np.linalg.norm(q, axis=1)).T
    if mu is not None:
        ang = np.dot(q, mu) < 0
        q[ang] = -q[ang]
    return q


def meanq(q, w=None):
    if w is None:
        return np.linalg.eigh(np.einsum('ij,ik->...jk', q, q))[1][:, -1]
    else:
        return np.linalg.eigh(np.einsum('ij,ik,i->...jk', q, q, w))[1][:, -1]


def pdistq(q1, q2=None):
    if q2 is None:
        q2 = q1
    d = np.abs(q1.dot(q2.T))
    # dots[dots > 1.0] = 1.0
    np.clip(d, 0, 1.0, out=d)
    np.arccos(d, out=d)
    d *= 2
    d_over_pi_half = d > np.pi/2
    d[d_over_pi_half] = np.pi - d[d_over_pi_half]
    d += 1e-6
    d **= 2
    d *= -0.5
    return d


def normdq(q, mu=None):
    mag = np.linalg.norm(q.real, axis=1)
    if mu is not None:
        ang = np.dot(q.real, mu) < 0
        q.real[ang] = -q.real[ang]
    return q / mag.reshape(q.shape[0], 1)




# Copyright (C) 2018 Daniel Asarnow
# University of California, San Francisco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from scipy.spatial import cKDTree


def bincorr(p1, p2, bins, minlength=0):
    bflat = bins.reshape(-1)
    p1flat = p1.reshape(-1)
    p2flat = p2.reshape(-1)
    fc = p1flat * np.conj(p2flat)
    fcr = np.bincount(bflat, fc.real, minlength=minlength)
    fcc = np.bincount(bflat, fc.imag, minlength=minlength)
    mag = np.sqrt(
        np.bincount(bflat, np.abs(p1flat) ** 2, minlength=minlength) *
        np.bincount(bflat, np.abs(p2flat) ** 2, minlength=minlength))
    mag[-1] += 1e-17
    frc = (fcr + fcc * 1j) / mag
    return frc


def query_connected(kdt, maxdist):
    if type(kdt) is not cKDTree:
        kdt = cKDTree(kdt)
    nb = np.full(kdt.n, np.nan)
    pairs = kdt.query_pairs(maxdist)
    for p in pairs:
        for idx in p[1:]:
            if np.isnan(nb[p[0]]):
                nb[idx] = p[0]
            else:
                nb[idx] = nb[p[0]]
    return nb




# Copyright (C) 2018 Daniel Asarnow
# University of California, San Francisco
from .util import *
from ..geom.convert import *
from ..geom.quat_numba import *
from ..geom.quat import normq
from ..geom.quat import meanq



# Copyright (C) 2018 Daniel Asarnow
# University of California, San Francisco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numba
import numpy as np
from .geom_numba import cross3_sca


@numba.jit(nopython=True, cache=False, nogil=True)
def _qconj(q, p):
    p[0] = q[0]
    p[1] = -q[1]
    p[2] = -q[2]
    p[3] = -q[3]
    return p


@numba.guvectorize(["void(float64[:], float64[:])"],
                   "(m)->(m)", nopython=True, cache=False)
def qconj(q, p):
    _qconj(q, p)


@numba.jit(nopython=True, cache=False, nogil=True)
def _qtimes(q1, q2, q3):
    q3[0] = q1[0] * q2[0] - (q1[1] * q2[1] + q1[2] * q2[2] + q1[3] * q2[3])
    q3[1] = q1[2] * q2[3] - q1[3] * q2[2] + q1[0] * q2[1] + q2[0] * q1[1]
    q3[2] = q1[3] * q2[1] - q1[1] * q2[3] + q1[0] * q2[2] + q2[0] * q1[2]
    q3[3] = q1[1] * q2[2] - q1[2] * q2[1] + q1[0] * q2[3] + q2[0] * q1[3]
    return q3


@numba.guvectorize(["void(float64[:], float64[:], float64[:])"],
                   "(m),(m)->(m)", nopython=True, cache=False)
def qtimes(q1, q2, q3):
    _qtimes(q1, q2, q3)


@numba.jit(nopython=True, cache=False, nogil=True)
def _qsqrt(q, p):
    p[0] = q[0] + 1
    p[1:] = q[1:]
    p[:] = p[:] / np.sqrt(2 * (1 + q[0]))
    return p


@numba.guvectorize(["void(float64[:], float64[:])"],
                   "(m)->(m)", nopython=True, cache=False)
def qsqrt(q, p):
    _qsqrt(q, p)


@numba.jit(cache=True, nopython=True, nogil=True)
def qslerp(q1, q2, t, longest=False):
    cos_half_theta = np.dot(q1, q2)
    if cos_half_theta >= 1.0:
        return q1.copy()
    if longest:
        if cos_half_theta > 0:
            cos_half_theta = -cos_half_theta
            q1 = -q1
    elif cos_half_theta < 0:
        cos_half_theta = -cos_half_theta
        q1 = -q1
    half_theta = np.arccos(cos_half_theta)
    sin_half_theta = np.sqrt(1 - cos_half_theta * cos_half_theta)
    if np.abs(sin_half_theta) < 1E-12:
        return (q1 + q2) / 2
    a = np.sin((1 - t) * half_theta)
    b = np.sin(t * half_theta)
    return (q1 * a + q2 * b) / sin_half_theta


@numba.jit(cache=False, nopython=True, nogil=True)
def distq(q1, q2):
    pi_half = np.pi / 2
    v = np.abs(np.sum(q1 * q2))
    v = np.clip(0, 1.0, v)
    v = np.arccos(v)
    v *= 2
    # msk = v > pi_half
    # v[msk] = np.pi - v[msk]
    return v


@numba.jit(cache=False, nopython=True, parallel=True)
def cdistq(q1, q2, d):
    # if q2 is None:
    #     q2 = q1
    # if out is None:
    # d = np.zeros((len(q1), len(q2)), dtype=q1.dtype)
    # else:
    #     d = out
    #     assert len(d) == len(q1)
    #     assert len(d[0]) == len(q2)
    pi_half = np.pi / 2
    for i in numba.prange(d.shape[0]):
        for j in range(d.shape[1]):
            v = np.abs(np.sum(q1[i] * q2[j]))
            if v > 1.0:
                v = 1.0
            v = np.arccos(v)
            v *= 2
            if v > pi_half:
                v = np.pi - v
            # v += 1e-6
            v **= 2
            v *= -0.5
            d[i, j] = v
    return d


@numba.jit(cache=False, nopython=True, parallel=True)
def pdistq(q1, d):
    pi_half = np.pi / 2
    for i in numba.prange(d.shape[0]):
        d[i, i] = 0.0
        for j in range(i + 1, d.shape[1]):
            v = np.abs(np.sum(q1[i] * q1[j]))
            if v > 1.0:
                v = 1.0
            v = np.arccos(v)
            v *= 2
            if v > pi_half:
                v = np.pi - v
            # v += 1e-6
            v **= 2
            v *= -0.5
            d[i, j] = d[j, i] = v
    return d


@numba.guvectorize(["void(complex128[:], complex128[:], complex128[:])"],
                   "(m),(m)->(m)", nopython=True, cache=False)
def dqtimes(q1, q2, q3):
    # Dual part = dual1 * real2 + real1 * dual2
    _qtimes(q1.imag, q2.real, q3.real)  # Store 1st dual term in real.
    _qtimes(q1.real, q2.imag, q3.imag)  # Store 2nd dual term in imag.
    tmp = q3.imag  # Workaround Numba typing error.
    tmp += q3.real  # Sum terms into imag.
    # Real part = real1 * real2
    _qtimes(q1.real, q2.real, q3.real)  # Overwrite real with real part.
    return


@numba.guvectorize(["void(complex128[:], complex128[:])"],
                   "(m)->(m)", nopython=True, cache=False)
def dqconj(q, p):
    _qconj(q.real, p.real)
    _qconj(q.imag, p.imag)


@numba.jit(nopython=True, cache=False, nogil=True)
def dqtimes_sca(q1, q2):
    q3 = np.zeros((4,), dtype=q1.dtype)
    # Dual part = dual1 * real2 + real1 * dual2
    _qtimes(q1.imag, q2.real, q3.real)  # Store 1st dual term in real.
    _qtimes(q1.real, q2.imag, q3.imag)  # Store 2nd dual term in imag.
    tmp = q3.imag  # Workaround Numba typing error.
    tmp += q3.real  # Sum terms into imag.
    # Real part = real1 * real2
    _qtimes(q1.real, q2.real, q3.real)  # Overwrite real with real part.
    return q3


@numba.jit(nopython=True, cache=False, nogil=True)
def dqconj_sca(q):
    p = np.zeros((4,), dtype=q.dtype)
    _qconj(q.real, p.real)
    _qconj(q.imag, p.imag)
    return p


@numba.jit(nopython=True, cache=False, nogil=True)
def dq2sc(q):
    theta = 2 * np.arccos(q[0].real)
    # nr = 1 / np.linalg.norm(q[1:].real)
    nr = np.sin(theta / 2)
    if nr < 1E-6:
        # nr = 0
        nr = 1e6
    else:
        nr = 1 / nr
    d = -2 * q[0].imag * nr
    l = q[1:].real * nr
    m = (q[1:].imag - l * d * q[0].real * 0.5) * nr
    return theta, d, l, m


@numba.jit(nopython=True, cache=False, nogil=True)
def sc2dq(theta, d, l, m):
    q = np.zeros((4,), dtype=np.complex128)
    theta_half = 0.5 * theta
    sin_theta_half = np.sin(theta_half)
    cos_theta_half = np.cos(theta_half)
    d_half = 0.5 * d
    q.real[0] = cos_theta_half
    q.real[:1] = sin_theta_half * l
    q.imag[0] = - d_half * sin_theta_half
    q.imag[1:] = sin_theta_half * m + d_half * cos_theta_half * l
    return q


# @numba.jit(nopython=True, cache=False, nogil=True)
# def dqpower(q, t):
#     theta, d, l, m = dq2sc(q)
#     theta_half = 0.5 * theta
#     d_half = 0.5 * d
#     da_real = t * np.cos(theta_half)
#     da_dual = t * -d_half * np.sin(theta_half)


@numba.jit(nopython=True, cache=False, parallel=True)
def pdistdq(q, d):
    relq = np.zeros((4,), dtype=np.complex128)
    for i in numba.prange(d.shape[0]):
        # relq = geom.dqtimes(geom.dqconj(q[i, None]), q)
        for j in range(i + 1, d.shape[1]):
            relq[:] = dqtimes_sca(dqconj_sca(q[i]), q[j])
            theta, dax, l, m = dq2sc(relq)
            p0 = cross3_sca(l, m)
            r2 = np.sum(p0 ** 2)
            v = np.sqrt(dax ** 2 + theta ** 2 * r2)
            d[i, j] = d[j, i] = v
    return d


@numba.jit(nopython=True, cache=False, parallel=True)
def cdistdq(q1, q2, d):
    relq = np.zeros((4,), dtype=np.complex128)
    for i in numba.prange(d.shape[0]):
        for j in range(d.shape[1]):
            relq = dqtimes_sca(dqconj_sca(q1[i]), q2[j])
            theta, dax, l, m = dq2sc(relq)
            p0 = cross3_sca(l, m)
            r2 = np.sum(p0 ** 2)
            v = np.sqrt(dax ** 2 + theta ** 2 * r2)
            d[i, j] = v
    return d


@numba.jit(nopython=True, cache=False, nogil=True)
def dqblend(q1, q2, t):
    q3 = (1 - t) * q1 + t * q2
    q3 /= np.linalg.norm(q3)
    return q3




# -*- coding: utf-8 -*-
# Copyright (C) 2018 Daniel Asarnow, Eugene Palovcak
# University of California, San Francisco
#
# Simple library for calculating contrast transfer function of electron microscopy.
# See help text and README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numba
import numpy as np


def ctf_freq(shape, d=1.0, full=True):
    """
    :param shape: Shape tuple.
    :param d: Frequency spacing in inverse Å (1 / pixel size).
    :param full: When false, return only unique Fourier half-space for real data. 
    """
    if full:
        xfrq = np.fft.fftfreq(shape[1])
    else:
        xfrq = np.fft.rfftfreq(shape[1])
    x, y = np.meshgrid(xfrq, np.fft.fftfreq(shape[0]))
    rho = np.sqrt(x**2 + y**2)
    a = np.arctan2(y, x)
    s = rho * d
    return s, a


@numba.jit(cache=True, nopython=True, nogil=True)
def eval_ctf(s, a, def1, def2, angast=0, phase=0, kv=300, ac=0.1, cs=2.0, bf=0, lp=0):
    """
    :param s: Precomputed frequency grid for CTF evaluation.
    :param a: Precomputed frequency grid angles.
    :param def1: 1st prinicipal underfocus distance (Å).
    :param def2: 2nd principal underfocus distance (Å).
    :param angast: Angle of astigmatism (deg) from x-axis to azimuth.
    :param phase: Phase shift (deg).
    :param kv:  Microscope acceleration potential (kV).
    :param ac:  Amplitude contrast in [0, 1.0].
    :param cs:  Spherical aberration (mm).
    :param bf:  B-factor, divided by 4 in exponential, lowpass positive.
    :param lp:  Hard low-pass filter (Å), should usually be Nyquist.
    """
    angast = np.deg2rad(angast)
    kv = kv * 1e3
    cs = cs * 1e7
    lamb = 12.2643247 / np.sqrt(kv * (1. + kv * 0.978466e-6))
    def_avg = -(def1 + def2) * 0.5
    def_dev = -(def1 - def2) * 0.5
    k1 = np.pi / 2. * 2 * lamb
    k2 = np.pi / 2. * cs * lamb**3
    k3 = np.sqrt(1 - ac**2)
    k4 = bf / 4.  # B-factor, follows RELION convention.
    k5 = np.deg2rad(phase)  # Phase shift.
    if lp != 0:  # Hard low- or high-pass.
        s *= s <= (1. / lp)
    s_2 = s**2
    s_4 = s_2**2
    dZ = def_avg + def_dev * (np.cos(2 * (a - angast)))
    gamma = (k1 * dZ * s_2) + (k2 * s_4) - k5
    ctf = -(k3 * np.sin(gamma) - ac*np.cos(gamma))
    if bf != 0:  # Enforce envelope.
        ctf *= np.exp(-k4 * s_2)
    return ctf


@numba.jit(cache=True, nopython=True, nogil=True)
def eval_ctf_between(n, apix, def1, def2, lores=0, hires=0, angast=0, phase=0, kv=300, ac=0.1, cs=2.0, bf=0, out=None):
    if out is None:
        out = np.zeros((n, n // 2 + 1))
    ctf = out.view()
    ctf.shape = (ctf.size,)
    d = 1. / (apix * n)
    kv = kv * 1e3
    cs = cs * 1e7
    lamb = 12.2643247 / np.sqrt(kv * (1. + kv * 0.978466e-6))
    def_avg = -(def1 + def2) * 0.5  # Sign convention for underfocused imaging.
    def_dev = -(def1 - def2) * 0.5
    k1 = np.pi / 2. * 2 * lamb
    k2 = np.pi / 2. * cs * lamb**3
    k3 = np.sqrt(1 - ac**2)
    k4 = bf / 4.  # B-factor, follows RELION convention.
    k5 = np.deg2rad(phase)  # Phase shift.
    rmin = int(lores * apix * n)
    rmin2 = rmin**2
    rmax = int(min(n // 2 - 1, hires * apix * n))
    rmax2 = rmax**2
    cnt = int(0)
    for i in range(n):
        if i <= rmax:
            yp = i
        elif i >= n - rmax:
            yp = i - n
        else:
            continue
        yp2 = yp ** 2
        for j in range(rmax + 1):
            xp = j
            r = xp ** 2 + yp2
            if r < rmin2 or r > rmax2:
                continue
            s = np.sqrt(xp**2 + yp**2) * d
            a = np.arctan2(yp, xp)
            s2 = s**2
            s4 = s**4
            dz = def_avg + def_dev * (np.cos(2 * (a - angast)))
            gamma = (k1 * dz * s2) + (k2 * s4) - k5
            # ctf[i, j] = -(k3 * np.sin(gamma) - ac * np.cos(gamma))
            ctf[cnt] = -(k3 * np.sin(gamma) - ac * np.cos(gamma))
            if bf != 0:  # Enforce envelope.
                # ctf[i, j] *= np.exp(-k4 * s2)
                ctf[cnt] *= np.exp(-k4 * s2)
            cnt += 1
    return out



# Copyright (C) 2017-2018 Daniel Asarnow
# University of California, San Francisco
#
# Library of miscellaneous utility functions.
# See help text and README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import
import bisect
import natsort
import numpy as np
import pandas as pd
import subprocess
from distutils.spawn import find_executable as which
from .. import geom
from .. import mrc
from .. import vop


def cent2edge(bins):
    """Convert bin centers to bin edges"""
    return np.r_[-np.Inf, 0.5 * (bins[:-1] + bins[1:]), np.Inf]


def relion_symmetry_group(sym):
    relion = which("relion_refine")
    if relion is None:
        raise RuntimeError(
            "Need relion_refine on PATH to obtain symmetry operators")
    stdout = subprocess.getoutput(
        "%s --sym %s --i /dev/null --o /dev/null --print_symmetry_ops" % (relion, sym))
    lines = stdout.split("\n")[2:]
    lines = [l for l in lines if not l.lstrip().startswith("Euler angles:")]
    return [np.array(
        [[np.double(val) for val in l.split()] for l in lines[i:i + 3]])
        for i in range(1, len(lines), 4)]


def aligndf(df1, df2, fields=None):
    ww = df1.set_index(fields)
    dd = df2.set_index(fields)
    i1 = set(tuple(f) for f in df1[fields].values)
    i2 = set(tuple(f) for f in df2[fields].values)
    iboth = list(i1.intersection(i2))
    df1a = ww.loc[iboth].copy()
    df2a = dd.loc[iboth].copy()
    return df1a, df2a


def interleave(dfs, drop=True):
    return pd.concat(dfs).sort_index(kind="mergesort").reset_index(drop=drop)


def join_struct_arrays(arrays):
    sizes = np.array([a.itemsize for a in arrays])
    offsets = np.r_[0, sizes.cumsum()]
    n = len(arrays[0])
    joint = np.empty((n, offsets[-1]), dtype=np.uint8)
    for a, size, offset in zip(arrays, sizes, offsets):
        joint[:, offset:offset + size] = a.view(np.uint8).reshape(n, size)
    dtype = sum((a.dtype.descr for a in arrays), [])
    return joint.ravel().view(dtype)


def dataframe_from_records_mapped(rec, field_dict):
    names = list(set([str(k) for k in field_dict if field_dict[k] is not None and k in rec.dtype.names]))
    df = pd.DataFrame.from_records(rec[names])
    df.columns = [field_dict[k] for k in names]
    return df


def nearest_good_box_size(n):
    b = [32, 36, 40, 48, 52, 56, 64, 66, 70, 72, 80, 84, 88, 100, 104, 108,
         112, 120, 128, 130, 132, 140, 144, 150, 160, 162, 168, 176, 180, 182,
         192, 200, 208, 216, 220, 224, 240, 256, 264, 288, 300, 308, 320, 324,
         336, 338, 352, 364, 384, 400, 420, 432, 448, 450, 462, 480, 486, 500,
         504, 512, 520, 528, 546, 560, 576, 588, 600, 640, 648, 650, 660, 672,
         686, 700, 702, 704, 720, 726, 728, 750, 768, 770, 784, 800, 810, 840,
         882, 896, 910, 924, 936, 972, 980, 1008, 1014, 1020, 1024, 1080, 1125,
         1152, 1200, 1215, 1250, 1280, 1296, 1350, 1440, 1458, 1500, 1536,
         1600, 1620, 1728, 1800, 1875, 1920, 1944, 2000, 2025, 2048, 2160,
         2187, 2250, 2304, 2400, 2430, 2500, 2560, 2592, 2700, 2880, 2916,
         3000, 3072, 3125, 3200, 3240, 3375, 3456, 3600, 3645, 3750, 3840,
         3888, 4000, 4050, 4320, 4374, 4500, 4608, 4800, 4860, 5000, 5120,
         5184, 5400, 5625, 5760, 5832, 6000, 6075, 6144, 6250, 6400, 6480,
         6750, 6912, 7200, 7290, 7500, 7680, 7776, 8000, 8100]
    return b[bisect.bisect(b, n) - 1]


def chimera_xform(xform, o=None, apix=1.):
    if o is None:
        o = np.array([0, 0, 0])
    r = xform[:, :3]
    v = xform[:, 3] / apix
    u = r.T.dot(o - v) - o
    return r, u


def chimera_xform2str(r, v):
    transform_string = np.column_stack([r, v]).tolist().__repr__()
    return transform_string


def chimera_xform2target(t0, r, u, o=None, apix=1.):
    if o is None:
        o = np.array([0, 0, 0])
    t1 = (r.dot(t0 / apix - o) + u + o) * apix
    return t1


def write_q_series(vol, qarr, basename, psz=1., order=1):
    for i, q in enumerate(qarr):
        r = geom.quat2rot(q / np.linalg.norm(q))
        decoy = vop.resample_volume(vol, r=r, order=order)
        mrc.write(basename % i, decoy, psz=psz)


def natsort_values(df, col, inplace=False):
    df = df if inplace else df.copy()
    if type(col) is str:
        idx = np.array(natsort.index_natsorted(df[col]))
    else:
        idx = np.array(natsort.index_natsorted(col))
    df["__natsort_key__"] = np.argsort(idx)
    df.sort_values("__natsort_key__", inplace=True)
    df.drop("__natsort_key__", axis=1, inplace=True)
    return df



# Copyright (C) 2017 Daniel Asarnow
# University of California, San Francisco
#
# Handles metadata from Frealign9, FrealignX, and cisTEM.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pandas as pd
from .. import star


def parse_f9_par(fn):
    head_data = {"Input particle images": None,
                 "Beam energy (keV)": None,
                 "Spherical aberration (mm)": None,
                 "Amplitude contrast": None,
                 "Pixel size of images (A)": None}
    ln = 1
    skip = 0
    with open(fn, 'rU') as f:
        lastheader = False
        firstblock = True
        for l in f:
            if l.startswith("C") and firstblock:
                if "PSI" in l or "DF1" in l:
                    lastheader = True
                    headers = l.rstrip().split()
                if ":" in l:
                    tok = l.split(":")
                    tok[1] = tok[1].lstrip().rstrip()
                    tok[0] = tok[0].lstrip("C ")
                    if tok[0] in head_data:
                        try:
                            head_data[tok[0]] = float(tok[1])
                        except ValueError:
                            head_data[tok[0]] = tok[1]
                if lastheader:
                    skip = ln
                    firstblock = False
            elif l.startswith("C"):
                break
            else:
                headers = ["C", "PHI", "THETA", "PSI", "SHX", "SHY",
                           "MAG", "FILM", "DF1", "DF2", "ANGAST",
                           "OCC", "LogP", "SIGMA", "SCORE", "CHANGE"]
                break

            ln += 1

    if skip == 0:
        n = None
    else:
        n = ln - skip - 1
    par = pd.read_table(fn, skiprows=skip, nrows=n, delimiter="\s+", header=None, comment="C")
    par.columns = headers
    for k in head_data:
        if head_data[k] is not None:
            par[k] = head_data[k]
    return par


def parse_fx_par(fn):
    with open(fn, 'r') as f:
        columns = f.readline().split()
        df = pd.read_csv(f, delimiter="\s+", header=None, names=columns, comment="C")
    return df


def write_f9_par(fn, df):
    formatters = {"C": lambda x: "%7d" % x,
                  "PSI": lambda x: "%7.2f" % x,
                  "THETA": lambda x: "%7.2f" % x,
                  "PHI": lambda x: "%7.2f" % x,
                  "SHX": lambda x: "%9.2f" % x,
                  "SHY": lambda x: "%9.2f" % x,
                  "MAG": lambda x: "%7.0f" % x,
                  "INCLUDE": lambda x: "%5d" % x,
                  "DF1": lambda x: "%8.1f" % x,
                  "DF2": lambda x: "%8.1f" % x,
                  "ANGAST": lambda x: "%7.2f" % x,
                  "PSHIFT": lambda x: "%7.2f" % x,
                  "OCC": lambda x: "%7.2f" % x,
                  "LogP": lambda x: "%9d" % x,
                  "SIGMA": lambda x: "%10.4f" % x,
                  "SCORE": lambda x: "%7.2f" % x,
                  "CHANGE": lambda x: "%7.2f" % x}
    with open(fn, 'w') as f:
        f.write(df.to_string(formatters=formatters, index=False))


def write_fx_par(fn, df):
    formatters = {"C": lambda x: "%7d" % x,
                  "PSI": lambda x: "%7.2f" % x,
                  "THETA": lambda x: "%7.2f" % x,
                  "PHI": lambda x: "%7.2f" % x,
                  "SHX": lambda x: "%9.2f" % x,
                  "SHY": lambda x: "%9.2f" % x,
                  "MAG": lambda x: "%7.0f" % x,
                  "INCLUDE": lambda x: "%5d" % x,
                  "DF1": lambda x: "%8.1f" % x,
                  "DF2": lambda x: "%8.1f" % x,
                  "ANGAST": lambda x: "%7.2f" % x,
                  "PSHIFT": lambda x: "%7.2f" % x,
                  "OCC": lambda x: "%7.2f" % x,
                  "LogP": lambda x: "%9d" % x,
                  "SIGMA": lambda x: "%10.4f" % x,
                  "SCORE": lambda x: "%7.2f" % x,
                  "CHANGE": lambda x: "%7.2f" % x}
    with open(fn, 'w') as f:
        f.write("C           PSI   THETA     PHI       SHX       SHY     "
                "MAG  INCLUDE   DF1      DF2  ANGAST  PSHIFT     "
                "OCC      LogP      SIGMA   SCORE  CHANGE\n")
        df.to_string(buf=f, formatters=formatters, index=False, header=None)
        f.write("\nC Blank line\n")


def par2star(par, data_path, apix=1.0, cs=2.0, ac=0.07, kv=300, invert_eulers=True):
    general = {"PHI": None,
            "THETA": None,
            "PSI": None,
            "SHX": None,
            "SHY": None,
            "MAG": None,
            "FILM": star.Relion.GROUPNUMBER,
            "DF1": star.Relion.DEFOCUSU,
            "DF2": star.Relion.DEFOCUSV,
            "ANGAST": star.Relion.DEFOCUSANGLE,
            "C": None,
            "CLASS": star.Relion.CLASS
            }
    rlnheaders = [general[h] for h in par.columns if h in general and general[h] is not None]
    df = par[[h for h in par.columns if h in general and general[h] is not None]].copy()
    df.columns = rlnheaders
    df[star.UCSF.IMAGE_INDEX] = par["C"] - 1
    df[star.UCSF.IMAGE_PATH] = data_path
    df[star.Relion.IMAGEPIXELSIZE] = apix
    df[star.Relion.CS] = cs
    df[star.Relion.AC] = ac
    df[star.Relion.VOLTAGE] = kv
    df[star.Relion.ORIGINXANGST] = -par["SHX"]
    df[star.Relion.ORIGINYANGST] = -par["SHY"]
    if invert_eulers:
        df[star.Relion.ANGLEROT] = -par["PSI"]
        df[star.Relion.ANGLETILT] = -par["THETA"]
        df[star.Relion.ANGLEPSI] = -par["PHI"]
    else:
        df[star.Relion.ANGLES] = par[["PHI", "THETA", "PSI"]]
    return df
